# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on binary file.
"""
import os
import os.path as path
import numpy as np
import shutil
from typing import Type, Iterable
from tqdm import tqdm
import logging
from kapture.utils.logging import getLogger
from kapture.utils.paths import path_secure
from enum import auto
from kapture.utils.Collections import AutoEnum

logger = getLogger()


# Binary data File IO ##################################################################################################
def array_from_file(filepath: str, dtype: Type, dsize: int) -> np.ndarray:
    """
    Reads a binary file of given type(dtype) and columns (dsize) into a numpy array.

    :param filepath: input path to binary file
    :param dtype: data type (eg. float)
    :param dsize: number of data (of type) per feature (eg. 2 for a simple keypoint)
    :return: numpy array
    """
    if not isinstance(dtype, type):
        raise TypeError('expect type as dtype.')
    if not isinstance(dsize, int) or dsize <= 0:
        raise TypeError('expect positive int as dsize.')

    with open(filepath, 'rb') as file:
        data_array = np.fromfile(file, dtype=dtype)
    data_array = data_array.reshape((-1, dsize))
    return data_array


def array_to_file(filepath: str, data_array: np.ndarray) -> None:
    """
    Writes the numpy array into a binary file.

    :param filepath:
    :param data_array:
    """
    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'wb') as f:
        data_array.tofile(f, sep='')


########################################################################################################################
class TransferAction(AutoEnum):
    """
    All possible operations when transferring recorded data (eg. images, Wifi, IMU, ...)
    or computed data (features, keypoints, ...).
    Transferring means importing, exporting, and more generally going from one representation
    to another representation of the data.
    """
    skip = auto()
    root_link = auto()
    copy = auto()
    move = auto()
    link_absolute = auto()
    link_relative = auto()


# transfer files functions
def transfer_files_from_dir_link(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        force_overwrite: bool = False,
        do_relative_link: bool = False
) -> None:
    """
    Transfer every files by linking given from the source list to destination list.
    The matching between source and kapture files are explicitly given.

    :param source_filepath_list: input list of source files. Uses guess_filepaths to obtains it from filenames.
    :param destination_filepath_list: input list of destination files (in kapture tree).
    :param force_overwrite: if True, overwrite destination file.
    :param do_relative_link: if True, do relative links else absolute links.
    """
    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    jobs = zip(source_filepath_list, destination_filepath_list)
    logger.debug('Linking data files')
    try:
        total_progress_bar = max(len(source_filepath_list), len(destination_filepath_list))
    except TypeError:
        total_progress_bar = None
    for src, dst in tqdm(jobs, disable=hide_progress_bar, total=total_progress_bar):
        # make sure we deal absolute full path
        src = path_secure(path.abspath(src))
        dst = path_secure(path.abspath(dst))
        os.makedirs(path.dirname(dst), exist_ok=True)
        if force_overwrite and path.lexists(dst):
            os.remove(dst)
        try:  # on windows, symlink requires some privileges, and may crash if not
            if do_relative_link:
                src = path.relpath(src, path.dirname(dst))
            os.symlink(src, dst)
        except OSError as e:
            logger.critical('unable to create symlink on image directory, due to privilege restrictions.')
            raise e


def transfer_files_from_dir_copy(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        force_overwrite: bool = False,
        delete_source: bool = False
) -> None:
    """
    Transfer every files by copying given from the source list to destination list.
    The matching between source and kapture files are explicitly given.
    If delete_source is activated, it moves files instead copying them.

    :param source_filepath_list: input list of absolute path to source files.
    :param destination_filepath_list: input list of absolute path to destination files.
    :param force_overwrite: if True, overwrite destination file.
    :param delete_source: if True, delete the imported files from source_record_dirpath.
    """
    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    logger.debug('Copying data files')
    for src, dst in tqdm(zip(source_filepath_list, destination_filepath_list), disable=hide_progress_bar):
        os.makedirs(path.dirname(dst), exist_ok=True)
        if force_overwrite and path.lexists(dst):
            os.remove(dst)
        if delete_source:
            shutil.move(src, dst)
        else:
            shutil.copyfile(src, dst)


def transfer_files_from_dir(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        copy_strategy: TransferAction = TransferAction.copy,
        force_overwrite: bool = False
) -> None:
    """
    Transfer files (copy or link files) from source to destination.
    The matching between source and kapture files are explicitly given.
    The actual import can be done using actual file copy (copy/move), or symlinks (absolute/relative).

    :param source_filepath_list: input list of source files.
    :param destination_filepath_list: input list of destination files (in kapture tree).
    :param copy_strategy: transfer strategy to apply.
    :param force_overwrite: if True, overwrite destination file.
    """
    if TransferAction.skip == copy_strategy:
        return

    elif TransferAction.copy == copy_strategy or TransferAction.move == copy_strategy:
        transfer_files_from_dir_copy(
            source_filepath_list=source_filepath_list,
            destination_filepath_list=destination_filepath_list,
            force_overwrite=force_overwrite,
            delete_source=(TransferAction.move == copy_strategy)
        )

    elif TransferAction.link_absolute == copy_strategy or TransferAction.link_relative == copy_strategy:
        transfer_files_from_dir_link(
            source_filepath_list=source_filepath_list,
            destination_filepath_list=destination_filepath_list,
            force_overwrite=force_overwrite,
            do_relative_link=(TransferAction.link_relative == copy_strategy)
        )

    else:
        raise ValueError(f'Unsupported copy action {copy_strategy}')



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license
# This script imports data from a EXIF


import argparse
import logging
import os.path as path
import path_to_kapture  # noqa: F401
import kapture
from kapture.converter.exif.import_exif import import_gps_from_exif

logger = logging.getLogger('exif')


def extract_exif_command_line() -> None:
    """
    Extract GPS coordinates from image EXIF metadata.
    """
    parser = argparse.ArgumentParser(description='Extract GPS coordinates from image EXIF metadata '
                                                 'and store in kapture files.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--input', '--kapture', required=True,
                        help='input path to kapture data root directory')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    args.input = path.abspath(args.input)
    logger.debug(''.join(['\n\t{:13} = {}'.format(k, v) for k, v in vars(args).items()]))
    # do the job
    import_gps_from_exif(kapture_dirpath=args.input)


if __name__ == '__main__':
    extract_exif_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on Feature related data.
Features are processed data linked to one or multiple record data.
For example Keypoints are features related to 1 RecordCamera (image).
And Matches are features related to a pair of RecordCamera (2 images).
"""

from kapture.io.tar import TarCollection, TarHandler, list_files_in_tar, retrieve_tar_handler_from_collection
import numpy as np
import os.path as path
from typing import Tuple, Any, Dict, Type, Optional, Union, Iterable

import kapture
from kapture.io.records import RECORD_DATA_DIRNAME
from kapture.utils.paths import path_secure, populate_files_in_dirpath
from kapture.io.binary import array_from_file, array_to_file


# Feature files related functions ######################################################################################
FEATURES_DATA_DIRNAMES = {
    kapture.Keypoints: path.join('reconstruction', 'keypoints'),
    kapture.Descriptors: path.join('reconstruction', 'descriptors'),
    kapture.GlobalFeatures: path.join('reconstruction', 'global_features'),
    kapture.Matches: path.join('reconstruction', 'matches'),
    kapture.RecordsBase: RECORD_DATA_DIRNAME,  # prefer kapture.RecordsCamera, Lidar, ...
}

FEATURE_FILE_EXTENSION = {
    kapture.Keypoints: '.kpt',
    kapture.Descriptors: '.desc',
    kapture.GlobalFeatures: '.gfeat',
    kapture.Matches: '.matches',
}

FEATURE_PAIR_PATH_SEPARATOR = {
    kapture.Matches: '.overlapping'
}


# get file path for binary files in kapture ############################################################################
def get_features_fullpath(
    data_type: Any,
    feature_type: str,
    kapture_dirpath: str = '',
    image_filename: Optional[str] = None,
    tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Union[str, Tuple[Optional[str], TarHandler]]:
    """
    Returns full path to subdirectory containing the binary files of the given type.
            Optionally, can give a the image file name, and add the feature file name (with proper extension).

    :param data_type:
    :param feature_type: name of the feature
    :param kapture_dirpath: input path to kapture directory.
    :param image_filename: optional input image filename (id).
    :param tar_handler: collection of preloaded tar archives
    :return: Feature full path
    """
    tar_local_handler = retrieve_tar_handler_from_collection(data_type, feature_type, tar_handler)
    if tar_local_handler is None:
        subdir = FEATURES_DATA_DIRNAMES[data_type]
        feature_filename = image_filename + FEATURE_FILE_EXTENSION[data_type] if image_filename else ''
        return path_secure(path.join(kapture_dirpath, subdir, feature_type, feature_filename))
    else:
        feature_filename = image_filename + FEATURE_FILE_EXTENSION[data_type] if image_filename else None
        return feature_filename, tar_local_handler


def features_to_filepaths(
    kapture_data: Union[kapture.Keypoints,
                        kapture.Descriptors,
                        kapture.GlobalFeatures],
    feature_type: str,
    kapture_dirpath: str = '',
    tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Dict[str, Union[str, Tuple[str, TarHandler]]]:
    """
    Returns a dict mapping image_id to path to feature file.
        Eg.{image_id: full/path/to/feature.jpg.kpt}
        or
        {image_id: (feature.jpg.kpt, ref_to_tar)}

    :param kapture_data: input kapture data.
    :param feature_type: name of the feature
    :param kapture_dirpath: input path to kapture directory.
    :param tar_handler: collection of preloaded tar archives
    :return: image id to image file path dictionary
    """
    data_type = type(kapture_data)
    return {
        image_filename: get_features_fullpath(data_type, feature_type, kapture_dirpath, image_filename, tar_handler)
        for image_filename in kapture_data
    }


def image_ids_from_feature_tar(kapture_type: Type[Union[kapture.Keypoints,
                                                        kapture.Descriptors,
                                                        kapture.GlobalFeatures]],
                               tar_handler: TarHandler) -> Iterable[str]:
    """
     Populate feature files in tar and returns their corresponding image_filename

    :param kapture_type: kapture class type.
    :param tar_handler: opened tar reference
    :return: image file paths
    """
    feature_filenames = list_files_in_tar(tar_handler, FEATURE_FILE_EXTENSION[kapture_type])
    image_filenames = (feature_filename[:-len(FEATURE_FILE_EXTENSION[kapture_type])]
                       for feature_filename in feature_filenames)
    return image_filenames


def image_ids_from_feature_dirpath(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures]],
        feature_type: str,
        kapture_dirpath: str = '') -> Iterable[str]:
    """
    Populate feature files and returns their corresponding image_filename.

    :param kapture_type: kapture class type.
    :param feature_type: name of the feature
    :param kapture_dirpath: root path of kapture
    :return: image file paths
    """
    feature_dirpath = get_features_fullpath(kapture_type, feature_type, kapture_dirpath)
    # filter only files with proper extension and
    feature_filenames = populate_files_in_dirpath(feature_dirpath, FEATURE_FILE_EXTENSION[kapture_type])
    # remove file extensions to retrieve image name.
    image_filenames = (feature_filename[:-len(FEATURE_FILE_EXTENSION[kapture_type])]
                       for feature_filename in feature_filenames)
    return image_filenames


def features_check_dir(
        kapture_data: Union[kapture.Keypoints,
                            kapture.Descriptors,
                            kapture.GlobalFeatures,
                            kapture.Matches],
        feature_type: str,
        kapture_dirpath: str,
        tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> bool:
    """
    Makes sure all files actually exists.

    :param kapture_data: input kapture data
    :param feature_type: name of the feature
    :param kapture_dirpath: root path of kapture
    :param tar_handler: collection of preloaded tar archives
    :return: True only if all exist, false otherwise
    """
    data_type = type(kapture_data)
    tar_local_handler = retrieve_tar_handler_from_collection(data_type, feature_type, tar_handler)
    file_list = features_to_filepaths(kapture_data, feature_type, kapture_dirpath, tar_local_handler).values()
    if tar_local_handler is None:
        all_files_exists = all(path.exists(feature_filepath) for feature_filepath in file_list)
    else:
        all_files_in_tar = set(list_files_in_tar(tar_local_handler, FEATURE_FILE_EXTENSION[data_type]))
        all_files_exists = all(feature_filepath[0] in all_files_in_tar for feature_filepath in file_list)
    return all_files_exists


# image_keypoints ######################################################################################################
def image_keypoints_from_file(filepath: Union[str, Tuple[str, TarHandler]], dtype: Type, dsize: int) -> np.ndarray:
    """
    Read the image keypoints

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image keypoints
    """
    if isinstance(filepath, str):
        return array_from_file(filepath, dtype, dsize)
    else:
        return filepath[1].get_array_from_tar(filepath[0], dtype, dsize)


def image_keypoints_to_file(filepath: Union[str, Tuple[str, TarHandler]], image_keypoints: np.ndarray) -> None:
    """
    Writes the image keypoints to file

    :param filepath: file path
    :param image_keypoints: image keypoints
    """
    if isinstance(filepath, str):
        array_to_file(filepath, image_keypoints)
    else:
        filepath[1].add_array_to_tar(filepath[0], image_keypoints)


def get_keypoints_fullpath(keypoints_type: str, kapture_dirpath: str, image_filename: Optional[str] = None,
                           tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Union[str, Tuple[Optional[str], TarHandler]]:
    """
    Computes the full path of the keypoints file

    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :param tar_handler: collection of preloaded tar archives
    :return: full path of the keypoints file
    """
    return get_features_fullpath(kapture.Keypoints, keypoints_type, kapture_dirpath, image_filename, tar_handler)


def keypoints_to_filepaths(keypoints: kapture.Keypoints, keypoints_type: str, kapture_dirpath: str,
                           tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Dict[str, Union[str, Tuple[str, TarHandler]]]:
    """
    Computes keypoints files paths

    :param keypoints: keypoints
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: keypoint to keypoint file dictionary
    """
    return features_to_filepaths(keypoints, keypoints_type, kapture_dirpath, tar_handler)


def keypoints_check_dir(keypoints: kapture.Keypoints, keypoints_type: str, kapture_dirpath: str,
                        tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> bool:
    """
    Checks that all keypoints file exist.

    :param keypoints: keypoints
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(keypoints, keypoints_type, kapture_dirpath, tar_handler)


# image_descriptors ####################################################################################################
def image_descriptors_from_file(filepath: Union[str, Tuple[str, TarHandler]], dtype: Type, dsize: int) -> np.ndarray:
    """
    Read the image descriptors

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image descriptors
    """
    if isinstance(filepath, str):
        return array_from_file(filepath, dtype, dsize)
    else:
        return filepath[1].get_array_from_tar(filepath[0], dtype, dsize)


def image_descriptors_to_file(filepath: Union[str, Tuple[str, TarHandler]], image_descriptors: np.ndarray) -> None:
    """
    Writes the image descriptors to file

    :param filepath: file path
    :param image_descriptors: image descriptors
    """
    if isinstance(filepath, str):
        array_to_file(filepath, image_descriptors)
    else:
        filepath[1].add_array_to_tar(filepath[0], image_descriptors)


def get_descriptors_fullpath(descriptors_type: str, kapture_dirpath: str, image_filename: Optional[str] = None,
                             tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Union[str, Tuple[Optional[str], TarHandler]]:
    """
    Computes the full path of the descriptors file

    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :param tar_handler: collection of preloaded tar archives
    :return: full path of the descriptors file
    """
    return get_features_fullpath(kapture.Descriptors, descriptors_type, kapture_dirpath, image_filename, tar_handler)


def descriptors_to_filepaths(descriptors: kapture.Descriptors,
                             descriptors_type: str,
                             kapture_dirpath: str,
                             tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Dict[str, Union[str, Tuple[str, TarHandler]]]:
    """
    Computes descriptors files paths

    :param descriptors: descriptors
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: descriptors to descriptors file dictionary
    """
    return features_to_filepaths(descriptors, descriptors_type, kapture_dirpath, tar_handler)


def descriptors_check_dir(descriptors: kapture.Descriptors, descriptors_type: str, kapture_dirpath: str,
                          tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> bool:
    """
    Checks that all descriptors file exist.

    :param descriptors: descriptors
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(descriptors, descriptors_type, kapture_dirpath, tar_handler)


# global_features ######################################################################################################
def image_global_features_from_file(filepath: Union[str, Tuple[str, TarHandler]], dtype: Type, dsize: int)\
        -> np.ndarray:
    """
    Read the image global features

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the global features
    """
    if isinstance(filepath, str):
        return array_from_file(filepath, dtype, dsize)
    else:
        return filepath[1].get_array_from_tar(filepath[0], dtype, dsize)


def image_global_features_to_file(filepath: Union[str, Tuple[str, TarHandler]],
                                  image_global_descriptor: np.ndarray) -> None:
    """
    Writes the image global features to file

    :param filepath: file path
    :param image_global_descriptor: image global features
    """
    if isinstance(filepath, str):
        array_to_file(filepath, image_global_descriptor)
    else:
        filepath[1].add_array_to_tar(filepath[0], image_global_descriptor)


def get_global_features_fullpath(global_features_type: str,
                                 kapture_dirpath: str,
                                 image_filename: Optional[str] = None,
                                 tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Union[str, Tuple[Optional[str], TarHandler]]:
    """
    Computes the full path of the global features file

    :param kapture_dirpath: top kapture directory path
    :param global_features_type: type of global_features, name of the global_features subfolder
    :param image_filename: image file name
    :param tar_handler: collection of preloaded tar archives
    :return: full path of the global features file
    """
    return get_features_fullpath(kapture.GlobalFeatures, global_features_type, kapture_dirpath,
                                 image_filename, tar_handler)


def global_features_to_filepaths(global_features: kapture.GlobalFeatures,
                                 global_features_type: str,
                                 kapture_dirpath: str,
                                 tar_handler: Optional[Union[TarCollection, TarHandler]] = None) \
        -> Dict[str, Union[str, Tuple[str, TarHandler]]]:
    """
    Computes global features files paths

    :param global_features: global features
    :param global_features_type: type of global_features, name of the global_features subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: global features to global features file dictionary
    """
    return features_to_filepaths(global_features, global_features_type, kapture_dirpath, tar_handler)


def global_features_check_dir(global_features: kapture.GlobalFeatures,
                              global_features_type: str,
                              kapture_dirpath: str,
                              tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> bool:
    """
    Checks that all keypoints file exist.

    :param global_features: global features
    :param global_features_type: type of global_features, name of the global_features subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(global_features, global_features_type, kapture_dirpath, tar_handler)


# matches ##############################################################################################################
def image_matches_from_file(filepath: Union[str, Tuple[str, TarHandler]]) -> np.ndarray:
    """
    Read the image matches
     [keypoint_id_image1, keypoint_id_image2, score]

    :param filepath: path to the file
    :return: matches
    """
    if isinstance(filepath, str):
        return array_from_file(filepath, dtype=np.float64, dsize=3)
    else:
        return filepath[1].get_array_from_tar(filepath[0], dtype=np.float64, dsize=3)


def image_matches_to_file(filepath: Union[str, Tuple[str, TarHandler]], image_matches: np.ndarray) -> None:
    """
    Writes the image matches to file

    :param filepath: file path
    :param image_matches: image matches
    """
    assert image_matches.dtype == np.float64
    assert image_matches.shape[1] == 3
    if isinstance(filepath, str):
        array_to_file(filepath, image_matches)
    else:
        filepath[1].add_array_to_tar(filepath[0], image_matches)


def get_matches_fullpath(
        image_filename_pair: Optional[Tuple[str, str]] = None,
        keypoints_type: str = '',
        kapture_dirpath: str = '',
        tar_handler: Optional[Union[TarCollection, TarHandler]] = None)  \
        -> Union[str, Tuple[Optional[str], TarHandler]]:
    """
    Computes the full path of the matches file between two images

    :param image_filename_pair: image file names
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: full path of the matches file
    """
    if image_filename_pair is not None:
        filename = path_secure(path.join(image_filename_pair[0] + FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches],
                                         image_filename_pair[1]))
    else:
        filename = None
    return get_features_fullpath(kapture.Matches, keypoints_type, kapture_dirpath, filename, tar_handler)


def matches_to_filepaths(matches: kapture.Matches,
                         keypoints_type: str,
                         kapture_dirpath: str = '',
                         tar_handler: Optional[Union[TarCollection, TarHandler]] = None)  \
        -> Dict[Tuple[str, str], Union[str, Tuple[str, TarHandler]]]:
    """
    Computes matches files paths

    :param matches: matches
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: file to file matches to matches file dictionary
    """
    return {
        image_filename_pair: get_matches_fullpath(image_filename_pair, keypoints_type, kapture_dirpath, tar_handler)
        for image_filename_pair in matches
    }


def _matches_filenames_remove_extensions_and_cut(matches_filenames: Iterable[str]) -> Iterable[Tuple[str, str]]:
    # remove the extensions and cut
    matching_pairs = (
        path_secure(matches_filename)[:-len(FEATURE_FILE_EXTENSION[kapture.Matches])
                                      ].split(FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches] + '/')
        for matches_filename in matches_filenames
    )
    matching_pairs = ((matches[0], matches[1]) for matches in matching_pairs if len(matches) == 2)
    return matching_pairs


def matching_pairs_from_tar(tar_handler: TarHandler) -> Iterable[Tuple[str, str]]:
    """
    Read and build Matches from a tar file.
    """
    matches_filenames = list_files_in_tar(tar_handler, FEATURE_FILE_EXTENSION[kapture.Matches])
    return _matches_filenames_remove_extensions_and_cut(matches_filenames)


def matching_pairs_from_dirpath(keypoints_type: str, kapture_dirpath: str) -> Iterable[Tuple[str, str]]:
    """
    Read and build Matches from kapture directory tree.
    """
    matches_dirpath = get_matches_fullpath(None, keypoints_type, kapture_dirpath)
    # list all files there is
    # filter only match files (the ones endings with .matches)
    matches_filenames = populate_files_in_dirpath(matches_dirpath, FEATURE_FILE_EXTENSION[kapture.Matches])
    return _matches_filenames_remove_extensions_and_cut(matches_filenames)


def matches_check_dir(matches: kapture.Matches, keypoints_type: str, kapture_dirpath: str,
                      tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> bool:
    """
    Checks that all matches file exist.

    :param matches: matches
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: top kapture directory path
    :param tar_handler: collection of preloaded tar archives
    :return: True if they all exist, false otherwise.
    """
    tar_local_handler = retrieve_tar_handler_from_collection(kapture.Matches, keypoints_type, tar_handler)
    file_list = (get_matches_fullpath(pair, keypoints_type, kapture_dirpath, tar_local_handler) for pair in matches)
    if tar_local_handler is None:
        all_files_exists = all(path.exists(filepath) for filepath in file_list)
    else:
        all_files_in_tar = set(list_files_in_tar(tar_local_handler, FEATURE_FILE_EXTENSION[kapture.Matches]))
        all_files_exists = all(feature_filepath[0] in all_files_in_tar for feature_filepath in file_list)
    return all_files_exists



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import os
import os.path as path
from random import randint
import numpy as np
from PIL import Image, ImageDraw

from .features import image_keypoints_from_file


def image_keypoints_to_image(
        image: Image,
        keypoints: np.ndarray,
        radius: int = 2,
        filled: bool = True) -> Image:
    """
    Displays keypoints on top of the image.

    :param image: an image
    :param keypoints: the keypoints
    :param radius: radius of the drawn circles
    :param filled: True: draw discs, False: draw circles
    :return: a new Image
    """
    draw = ImageDraw.Draw(image)
    for coords in keypoints[:, 0:2]:
        coords_tl = (coords - radius).astype(int)
        coords_br = (coords + radius).astype(int)
        color = tuple([randint(0, 255) for _ in range(3)])
        draw.ellipse((coords_tl[0], coords_tl[1], coords_br[0], coords_br[1]), outline=color,
                     fill=color if filled else None)
    return image


def image_keypoints_to_image_file(
        output_filepath: str,
        image_filepath: str,
        keypoints_filepath: str,
        keypoint_dtype,
        keypoint_dsize,
        radius: int = 2) -> None:
    """
    Displays keypoints on top of the image and save it to image file.

    :param output_filepath: input path to output image of keypoints.
    :param image_filepath: input path to input image.
    :param keypoints_filepath: input path to keypoints file.
    :param keypoint_dtype: input data type of keypoints data (cf. binary).
    :param keypoint_dsize: input data size of keypoints data (cf. binary)
    :param radius: radius of the keypoint in image (in pixel).
    :return:
    """
    os.makedirs(path.dirname(output_filepath), exist_ok=True)
    keypoints = image_keypoints_from_file(keypoints_filepath, keypoint_dtype, keypoint_dsize)
    image = Image.open(image_filepath)
    image_of_keypoints = image_keypoints_to_image(image, keypoints, radius)
    image_of_keypoints.save(output_filepath)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Colmap camera management
"""

import kapture


CAMERA_MODEL_NAME_ID = [
    ('SIMPLE_PINHOLE', 0),
    ('PINHOLE', 1),
    ('SIMPLE_RADIAL', 2),
    ('RADIAL', 3),
    ('OPENCV', 4),
    ('OPENCV_FISHEYE', 5),
    ('FULL_OPENCV', 6),
    ('FOV', 7),
    ('SIMPLE_RADIAL_FISHEYE', 8),
    ('RADIAL_FISHEYE', 9),
    ('THIN_PRISM_FISHEYE', 10),
]

CAMERA_MODEL_NAMES = {i: name for name, i in CAMERA_MODEL_NAME_ID}
CAMERA_MODEL_IDS = {name: i for name, i in CAMERA_MODEL_NAME_ID}

DEFAULT_FOCAL_LENGTH_FACTOR = 1.2


def get_camera_kapture_id_from_colmap_id(camera_id_colmap) -> str:
    """
    Create a deterministic kapture camera identifier from the colmap camera identifier:
    sensor_id = "cam_xxxxx"  where "xxxxx" is the colmap ID.

    :param camera_id_colmap: colmap camera identifier
    :return: kapture camera identifier.
    """
    assert isinstance(camera_id_colmap, int)
    return f'cam_{camera_id_colmap:05d}'


def get_colmap_camera(camera: kapture.Camera):
    """
    Compute the colmap camera definition.

    :param camera: a kapture camera definition
    :return: colmap camera parameters.
    """
    assert isinstance(camera, kapture.Camera)
    assert len(camera.camera_params) >= 2

    width = camera.camera_params[0]
    height = camera.camera_params[1]
    if camera.camera_type == kapture.CameraType.UNKNOWN_CAMERA:
        prior_focal_length = False
        camera_id = CAMERA_MODEL_IDS[kapture.CameraType.SIMPLE_RADIAL.value]
        focal_length = DEFAULT_FOCAL_LENGTH_FACTOR * max(width, height)
        cx = width/2.0
        cy = height/2.0
        params = [focal_length, cx, cy, 0.0]
    elif camera.camera_type.value in CAMERA_MODEL_IDS:
        prior_focal_length = True
        camera_id = CAMERA_MODEL_IDS[camera.camera_type.value]
        params = camera.camera_params[2:]
    else:
        raise ValueError(
            'This sensor model: {} is not supported by colmap'.format(camera.camera_type.value))
    return camera_id, width, height, params, prior_focal_length



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Virtual gallery testing data import.
"""

from collections import OrderedDict
import logging
import quaternion
from typing import Iterable, List
# kapture
import kapture
# local
from .virtual_gallery_constants import virtual_gallery_camera_model, virtual_gallery_width, virtual_gallery_height

logger = logging.getLogger('virtual_gallery')


class VirtualGalleryTestingIntrinsic:
    """
    virtual gallery testing intrinsics
    """

    def __init__(self, light_id: int, occlusion_id: int, frame_id: int, intrinsics: list):
        self.light_id = light_id
        self.occlusion_id = occlusion_id
        self.frame_id = frame_id
        self.intrinsics = intrinsics

    def __hash__(self):
        return hash((self.light_id, self.occlusion_id,  self.frame_id, tuple(self.intrinsics)))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.light_id == other.light_id and
                self.occlusion_id == other.occlusion_id and
                self.frame_id == other.camera_id and
                self.frame_id == other.intrinsics)


class VirtualGalleryTestingExtrinsic:
    """
    virtual gallery testing extrinsics
    """

    def __init__(self, frame_id: int, light_id: int, occlusion_id: int, extrinsics: list):
        self.light_id = light_id
        self.extrinsics = extrinsics
        self.frame_id = frame_id
        self.occlusion_id = occlusion_id

    def __hash__(self):
        return hash((self.light_id, self.occlusion_id, tuple(self.extrinsics), self.frame_id))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.light_id == other.light_id and
                self.occlusion_id == other.occlusion_id and
                self.extrinsics == other.extrinsics and
                self.frame_id == other.frame_id)


def import_testing_intrinsics(input_root: str,
                              light_range: list,
                              occlusion_range: list
                              ) -> List[VirtualGalleryTestingIntrinsic]:
    """
    Import testing intrinsics
    # format of intrinsic.txt is
    # frame cameraID K[0,0] K[1,1] K[0,2] K[1,2]
    # 0 0 1371.022 1371.022 959.5 539.5

    :param input_root:
    :param light_range:
    :param occlusion_range:
    :return: intrinsics
    """
    # assert that all intrinsics are identical for one camera and take any

    intrinsics = []
    for light_id in light_range:
        for occlusion_id in occlusion_range:
            with open(f'{input_root}/testing/gallery_light{light_id}'
                      f'_occlusion{occlusion_id}/intrinsic.txt', 'r') as file:
                lines = file.readlines()
                for split in map(lambda x: x.split(), lines):
                    if split[0].isdigit():
                        intrinsic = VirtualGalleryTestingIntrinsic(
                            light_id, occlusion_id, int(split[0]),
                            [float(split[2]), float(split[3]), float(split[4]), float(split[5])]
                        )
                        intrinsics.append(intrinsic)
    return intrinsics


def import_testing_extrinsics(input_root: str,
                              light_range: list,
                              occlusion_range: list
                              ) -> List[VirtualGalleryTestingExtrinsic]:
    """
    Import testing extrinsics.

    # format of extrinsic.txt is
    # frame cameraID r1,1 r1,2 r1,3 t1 r2,1 r2,2 r2,3 t2 r3,1 r3,2 r3,3 t3 0 0
    # 0 1
    # 0 0 0.929348 0 0.3692049 2.798107 0 1 0 1.65 -0.3692049 0 0.929348
    # 1.618328 0 0 0 1

    :param input_root:
    :param light_range:
    :param occlusion_range:
    :return: extrinsics
    """
    extrinsics = []
    for light_id in light_range:
        for occlusion_id in occlusion_range:
            with open(f'{input_root}/testing/gallery_light{light_id}'
                      f'_occlusion{occlusion_id}/extrinsic.txt', 'r') as file:
                lines = file.readlines()
                for split in map(lambda x: x.split(), lines):
                    if split[0].isdigit():
                        extrinsic = VirtualGalleryTestingExtrinsic(
                            int(split[0]), light_id, occlusion_id,
                            [float(value) for value in split[2:]]
                        )
                        extrinsics.append(extrinsic)
    return extrinsics


def _get_testing_camera_name(light_id: int, occlusion_id: int, frame_id: int) -> str:
    return f'testing_light_{light_id}_occlusion_{occlusion_id}_frame_{frame_id}'


def convert_testing_intrinsics(testing_intrinsics: Iterable[VirtualGalleryTestingIntrinsic],
                               sensors: kapture.Sensors) -> None:
    """
    Import all testing intrinsics into the sensors definitions.

    :param testing_intrinsics: testing intrinsics to import
    :param sensors: list of sensor definitions where to add the new definitions
    """
    logger.info("Converting testing cameras...")
    for intrinsic in testing_intrinsics:
        camera_device_id = _get_testing_camera_name(intrinsic.light_id, intrinsic.occlusion_id, intrinsic.frame_id)
        camera = kapture.Camera(virtual_gallery_camera_model,
                                [virtual_gallery_width, virtual_gallery_height] + intrinsic.intrinsics)
        sensors[camera_device_id] = camera


def convert_testing_extrinsics(offset: int,
                               testing_extrinsics: Iterable[VirtualGalleryTestingExtrinsic],
                               images: kapture.RecordsCamera,
                               trajectories: kapture.Trajectories) -> None:
    """
    Import all testing extrinsics into the images and trajectories.

    :param offset:
    :param testing_extrinsics: testing extrinsics to import
    :param images: image list to add to
    :param trajectories: trajectories to add to
    """
    # Map (light_id, loop_id, frame_id) to a unique timestamp
    testing_frames_tuples = ((extrinsic.light_id, extrinsic.occlusion_id, extrinsic.frame_id)
                             for extrinsic in testing_extrinsics)
    testing_frames_tuples = OrderedDict.fromkeys(testing_frames_tuples).keys()
    testing_frame_mapping = {v: n + offset for n, v in enumerate(testing_frames_tuples)}

    # Export images and trajectories
    logger.info("Converting testing images and trajectories...")
    for extrinsic in testing_extrinsics:
        rotation_matrix = [extrinsic.extrinsics[0:3], extrinsic.extrinsics[4:7], extrinsic.extrinsics[8:11]]
        rotation = quaternion.from_rotation_matrix(rotation_matrix)
        timestamp = testing_frame_mapping[(extrinsic.light_id, extrinsic.occlusion_id, extrinsic.frame_id)]
        camera_device_id = _get_testing_camera_name(extrinsic.light_id, extrinsic.occlusion_id, extrinsic.frame_id)
        translation_vector = [extrinsic.extrinsics[3], extrinsic.extrinsics[7], extrinsic.extrinsics[11]]
        images[(timestamp, camera_device_id)] = (f'testing/gallery_light{extrinsic.light_id}_occlusion'
                                                 f'{extrinsic.occlusion_id}/frames/rgb/'
                                                 f'camera_0/rgb_{extrinsic.frame_id:05}.jpg')
        trajectories[(timestamp, camera_device_id)] = kapture.PoseTransform(rotation, translation_vector)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import os
import tarfile
import hashlib
import logging

logger = logging.getLogger('downloader')


def untar_file(archive_filepath: str,
               install_dirpath: str):
    """
    Equivalent to tar -xf <archive_filepath> -C <install_dirpath>

    :param archive_filepath: input full path to the archive file.
    :param install_dirpath: input full path to directory where to extract.
    """
    logger.debug(f'extracting\n\tfrom: {archive_filepath}\n\tto  : {install_dirpath}')
    # make sure directory exists
    os.makedirs(install_dirpath, exist_ok=True)
    with tarfile.open(archive_filepath, 'r:*') as archive:
        archive.extractall(install_dirpath)


def compute_sha256sum(archive_filepath: str):
    """
    Computes the sha256sum on the given file.

    :param archive_filepath: input full path to file.
    :return the sha256sum.
    """
    sha256_hash = hashlib.sha256()
    with open(archive_filepath, 'rb') as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports data from a COLMAP database and/or reconstruction files

 It will assume that if all rotation parameters are null, there is no prior rotation, and if all translation parameters
 are null, there is no prior translation (using null values instead of zeros)
"""

import logging
import os
import os.path as path
from typing import Optional, Set, Type, Union

# kapture
import kapture
import kapture.io.features
import kapture.io.records
import kapture.io.structure
import kapture.algo.merge_keep_ids
from kapture.io.records import TransferAction, import_record_data_from_dir_auto

# local
from .database import COLMAPDatabase
from .import_colmap_database import get_cameras_from_database, get_images_and_trajectories_from_database
from .import_colmap_database import get_keypoints_from_database, get_descriptors_from_database
from .import_colmap_database import get_matches_from_database
from .import_colmap_rigs import import_colmap_rig
from .import_colmap_reconstruction import import_from_colmap_cameras_txt, import_from_colmap_images_txt
from .import_colmap_reconstruction import import_from_colmap_points3d_txt

logger = logging.getLogger('colmap')


def import_colmap_database(colmap_database_filepath: str,
                           kapture_dir_path: Optional[str],
                           keypoints_type: str = 'SIFT',
                           descriptors_type: str = 'SIFT',
                           no_geometric_filtering: bool = False,
                           skip_reconstruction: bool = False) -> kapture.Kapture:
    """
    Converts colmap database file to kapture data.
    If kapture_dir_path is given, it creates keypoints, descriptors, matches files (if any).

    :param colmap_database_filepath: path to colmap database file.
    :param kapture_dir_path: path to kapture directory. Is used to store keypoints, descriptors and matches files.
                            If not given (None), is equivalent to skip_reconstruction == True.
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param no_geometric_filtering:
    :param skip_reconstruction: skip the import of the kapture/reconstruction part,
                            ie. Keypoints, Descriptors, Matches.
    :return: kapture object
    """
    kapture_data = kapture.Kapture()

    logger.debug(f'loading colmap database {colmap_database_filepath}')
    db = COLMAPDatabase.connect(colmap_database_filepath)

    # Generate cameras
    logger.debug('parsing cameras in database.')
    kapture_data.sensors = get_cameras_from_database(db)

    # Images, Trajectories
    logger.debug('parsing images and trajectories in database.')
    kapture_data.records_camera, kapture_data.trajectories = get_images_and_trajectories_from_database(db)

    if kapture_dir_path is not None and not skip_reconstruction:
        os.makedirs(kapture_dir_path, exist_ok=True)

        # keypoints
        logger.debug('parsing keypoints in database...')
        keypoints = get_keypoints_from_database(db, kapture_data.records_camera, kapture_dir_path, None, keypoints_type)
        if keypoints is not None:
            kapture_data.keypoints = {keypoints_type: keypoints}

        # descriptors
        logger.debug('parsing descriptors in database...')
        descriptors = get_descriptors_from_database(db, kapture_data.records_camera, kapture_dir_path,
                                                    None, keypoints_type, descriptors_type)
        if descriptors is not None:
            kapture_data.descriptors = {descriptors_type: descriptors}

        # matches
        logger.debug('parsing matches in database...')
        matches = get_matches_from_database(db, kapture_data.records_camera, kapture_dir_path,
                                            None, keypoints_type, no_geometric_filtering)
        if matches is not None:
            kapture_data.matches = {keypoints_type: matches}

    db.close()
    return kapture_data


def import_colmap_from_reconstruction_files(reconstruction_dir_path: str,
                                            kapture_dir_path: Optional[str],
                                            keypoints_type: str = 'SIFT',
                                            skip: Set[Type[Union[kapture.Keypoints,
                                                                 kapture.Points3d,
                                                                 kapture.Observations]]] = set()
                                            ) -> kapture.Kapture:
    """
    Converts colmap reconstruction files to kapture data.
    If kapture_dir_path is given, keypoints files are created, and potentially their observations.

    :param reconstruction_dir_path:
    :param kapture_dir_path: path to kapture directory. Is used to store keypoints files.
                            If not given (None), keypoints are automatically skipped.
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param skip: can skip independently : Keypoints, Points3d or Observations.
                Note that Points3d and Observations are in the same file, so you should skip both to gain its reading.
    :return: kapture object
    """
    logger.debug(f'loading colmap reconstruction from:\n\t"{reconstruction_dir_path}"')
    if skip:
        logger.debug(f'loading colmap reconstruction skipping {", ".join(s.__name__ for s in skip)}')

    kapture_data = kapture.Kapture()
    reconstruction_file_paths = (path.join(reconstruction_dir_path, filename)
                                 for filename in ['cameras.txt', 'images.txt', 'points3D.txt'])
    colmap_cameras_filepath, colmap_images_filepath, colmap_points3d_filepath = reconstruction_file_paths

    proceed_keypoints = kapture.Keypoints not in skip and kapture_dir_path is not None
    proceed_points3d = kapture.Points3d not in skip and path.exists(colmap_points3d_filepath)
    proceed_observations = kapture.Observations not in skip and path.exists(colmap_points3d_filepath)

    if path.exists(colmap_cameras_filepath):
        logging.debug(f'parsing cameras from:\n\t"{path.basename(colmap_cameras_filepath)}"')
        kapture_data.sensors = import_from_colmap_cameras_txt(colmap_cameras_filepath)

    if path.exists(colmap_images_filepath):
        logging.debug(f'loading images from:\n\t"{path.basename(colmap_images_filepath)}"')
        kapture_dir_path_for_keypoints = kapture_dir_path if proceed_keypoints else None
        images, trajectories, keypoints = import_from_colmap_images_txt(
            colmap_images_filepath, kapture_dir_path_for_keypoints, keypoints_type)

        kapture_data.records_camera = images
        kapture_data.trajectories = trajectories
        if keypoints is not None:
            if kapture_data.keypoints is None:
                kapture_data.keypoints = {}
            kapture_data.keypoints[keypoints_type] = keypoints

    if proceed_points3d or proceed_observations:
        assert kapture_data.records_camera is not None
        image_id_2_names = {ts: image_name
                            for ts, cam_id, image_name in kapture.flatten(kapture_data.records_camera, True)}
        logger.debug(f'parsing 3d points and observations from:\n\t"{path.basename(colmap_points3d_filepath)}"')
        points3d, observations = import_from_colmap_points3d_txt(colmap_points3d_filepath,
                                                                 image_id_2_names,
                                                                 keypoints_type,
                                                                 skip_observations=False)
        kapture_data.points3d = points3d if proceed_points3d else None
        kapture_data.observations = observations if proceed_observations else None

    return kapture_data


def import_colmap(kapture_dir_path: Optional[str],  # noqa: C901: the import algorithm is well documented
                  colmap_database_filepath: str = None,
                  colmap_reconstruction_dir_path: str = None,
                  colmap_images_dir_path: str = None,
                  colmap_rig_filepath: str = None,
                  keypoints_type: str = 'SIFT',
                  descriptors_type: str = 'SIFT',
                  no_geometric_filtering: bool = False,
                  skip_reconstruction: bool = False,
                  force_overwrite_existing: bool = False,
                  images_import_strategy: TransferAction = TransferAction.link_absolute
                  ) -> kapture.Kapture:
    """
    Converts colmap files to kapture object.

    :param kapture_dir_path: path to kapture directory. Is used to store keypoints, descriptors and matches files.
                            If not given (None), keypoints, descriptors and matches are skipped.
    :param colmap_database_filepath: optional path to colmap database file.
    :param colmap_reconstruction_dir_path: optional path to colmap reconstruction directory.
    :param colmap_images_dir_path: directory path to colmap images. If given, a link to it will be created.
    :param colmap_rig_filepath: optional path to colmap rig file.
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
     (by default, in colmap, it is SIFT, but can be imported)
    :param no_geometric_filtering:
    :param skip_reconstruction: skip the import of the kapture/reconstruction part,
                                ie. Keypoints, Descriptors, Matches, Points3d, Observations.
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_strategy: input choice: how to copy image files.
    :return: kapture object
    """

    # sanity checks
    if kapture_dir_path and colmap_images_dir_path and images_import_strategy == TransferAction.skip:
        logger.warning('Images from colmap will not be copied (skip).')

    # prepare output directory
    if kapture_dir_path:
        kapture.io.structure.delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)
        os.makedirs(kapture_dir_path, exist_ok=True)

    # 1: import database
    kapture_from_database = None
    if colmap_database_filepath:
        logger.debug(f'importing from database "{colmap_database_filepath}"')
        kapture_from_database = import_colmap_database(
            colmap_database_filepath, kapture_dir_path,
            keypoints_type, descriptors_type,
            no_geometric_filtering, skip_reconstruction)

    # 2: import reconstruction text files.
    kapture_data_reconstructed = None
    if colmap_reconstruction_dir_path:
        # do not overwrite keypoints files if any from database import
        what_to_skip_during_import_txt = set()
        # if keypoints already loaded from DB,
        # do not load them again and overwrite them
        # because keypoints from txt have less info:
        if kapture_from_database and kapture_from_database.keypoints:
            what_to_skip_during_import_txt.add(kapture.Keypoints)
        # skip_reconstruction=skip keypoints, Points3d, Observations
        if skip_reconstruction:
            what_to_skip_during_import_txt |= {kapture.Keypoints, kapture.Points3d, kapture.Observations}
        logger.debug(f'importing from reconstruction "{colmap_reconstruction_dir_path}"')
        kapture_data_reconstructed = import_colmap_from_reconstruction_files(
            colmap_reconstruction_dir_path, kapture_dir_path, keypoints_type, what_to_skip_during_import_txt)

    # Merge data from database and reconstruction files
    if colmap_database_filepath and colmap_reconstruction_dir_path:
        # if both are present:
        # - kapture_data.sensors: merge both, with priority to reconstruction.
        # - kapture_data.trajectories: keep only reconstruction trajectories.
        # - kapture_data.observations: keep only reconstruction observations.
        # - kapture_data.points3d: only exists in colmap reconstruction
        # - kapture_data.*: anything else, keep only database

        # by default take all from database
        kapture_data = kapture_from_database
        # just replace trajectories, observations, points3d
        kapture_data.trajectories = kapture_data_reconstructed.trajectories
        kapture_data.observations = kapture_data_reconstructed.observations
        kapture_data.points3d = kapture_data_reconstructed.points3d
        # do a merge for sensors. If conflict prefer reconstruction:
        kapture_data.sensors.update(kapture_data_reconstructed.sensors)

    elif colmap_database_filepath:
        kapture_data = kapture_from_database
    elif colmap_reconstruction_dir_path:
        kapture_data = kapture_data_reconstructed
    else:
        raise ValueError('Neither database nor reconstruction files where given.')

    # if there is a rig ! lets restore it, and restore also timestamps
    if colmap_rig_filepath:
        rigs, records_camera, trajectories = import_colmap_rig(
            colmap_rig_filepath,
            kapture_data.records_camera,
            kapture_data.trajectories)

        kapture_data.rigs = rigs
        if records_camera:
            if not len(list(kapture.flatten(kapture_data.records_camera))) == len(
                    list(kapture.flatten(records_camera))):
                raise ValueError('inconsistent timestamp reconstruction in images')
            kapture_data.records_camera = records_camera

        if trajectories:
            if not len(list(kapture.flatten(kapture_data.trajectories))) == len(list(kapture.flatten(trajectories))):
                raise ValueError('inconsistent timestamp reconstruction in trajectories')
            kapture_data.trajectories = trajectories

    # finally import images
    if kapture_dir_path and colmap_images_dir_path and images_import_strategy != TransferAction.skip:
        filename_list = [f for _, _, f in kapture.flatten(kapture_data.records_camera)]
        logger.info(f'importing {len(filename_list)} image files ...')
        import_record_data_from_dir_auto(
            colmap_images_dir_path,
            kapture_dir_path,
            filename_list,
            images_import_strategy
        )
    assert kapture_data is not None
    return kapture_data



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
All reading and writing operations of kapture objects in CSV like files
"""

from collections import namedtuple
import datetime
import io
import numpy as np
import quaternion
import os
import os.path as path
import re
import sys
from typing import Any, Dict, List, Optional, Set, Type, Union


import kapture
import kapture.io.features
from kapture.utils.logging import getLogger
from .tar import KAPTURE_TARABLE_TYPES, TarCollection, TarHandler
from .tar import get_feature_tar_fullpath, retrieve_tar_handler_from_collection

logger = kapture.logger
gettrace = getattr(sys, 'gettrace', None)

# file names conventions
CSV_FILENAMES = {
    kapture.Sensors: path.join('sensors', 'sensors.txt'),
    kapture.Trajectories: path.join('sensors', 'trajectories.txt'),
    kapture.Rigs: path.join('sensors', 'rigs.txt'),
    kapture.RecordsCamera: path.join('sensors', 'records_camera.txt'),
    kapture.RecordsDepth: path.join('sensors', 'records_depth.txt'),
    kapture.RecordsLidar: path.join('sensors', 'records_lidar.txt'),
    kapture.RecordsWifi: path.join('sensors', 'records_wifi.txt'),
    kapture.RecordsBluetooth: path.join('sensors', 'records_bluetooth.txt'),
    kapture.RecordsGnss: path.join('sensors', 'records_gnss.txt'),
    kapture.RecordsAccelerometer: path.join('sensors', 'records_accelerometer.txt'),
    kapture.RecordsGyroscope: path.join('sensors', 'records_gyroscope.txt'),
    kapture.RecordsMagnetic: path.join('sensors', 'records_magnetic.txt'),
    kapture.Points3d: path.join('reconstruction', 'points3d.txt'),
    kapture.Observations: path.join('reconstruction', 'observations.txt'),
}

FEATURES_CSV_FILENAMES = {
    kapture.Keypoints: lambda x: path.join('reconstruction', 'keypoints', x, 'keypoints.txt'),
    kapture.Descriptors: lambda x: path.join('reconstruction', 'descriptors', x, 'descriptors.txt'),
    kapture.GlobalFeatures: lambda x: path.join('reconstruction', 'global_features', x, 'global_features.txt'),
}


def get_csv_fullpath(kapture_type: Any, kapture_dirpath: str = '') -> str:
    """
    Returns the full path to csv kapture file for a given data structure and root directory.
    This path is the concatenation of the kapture root path and subpath into kapture into data structure.

    :param kapture_type: type of kapture data (kapture.RecordsCamera, kapture.Trajectories, ...)
    :param kapture_dirpath: root kapture path
    :return: full path of csv file for that type of data
    """
    assert kapture_type in CSV_FILENAMES
    filename = CSV_FILENAMES[kapture_type]
    return path.join(kapture_dirpath, filename)


def get_feature_csv_fullpath(kapture_type: Any, feature_name: str, kapture_dirpath: str = '') -> str:
    """
    Returns the full path to csv kapture file for a given datastructure and root directory.
    This path is the concatenation of the kapture root path and subpath into kapture into data structure.

    :param kapture_type: type of kapture data (kapture.RecordsCamera, kapture.Trajectories, ...)
    :param feature_name: type of keypoints, descriptors, global_features, used to infer the path...
    :param kapture_dirpath: root kapture path
    :return: full path of csv file for that type of data
    """
    assert kapture_type in FEATURES_CSV_FILENAMES
    filename = FEATURES_CSV_FILENAMES[kapture_type](feature_name)
    return path.join(kapture_dirpath, filename)


PADDINGS = {
    'timestamp': [8],
    'device_id': [3],
    'pose': [4, 4, 4, 4, 4, 4, 4],
}

KAPTURE_FORMAT_1 = "# kapture format: 1.1"
KAPTURE_FORMAT_PARSING_RE = '# kapture format\\:\\s*(?P<version>\\d+\\.\\d+)'

# Line separator for the kapture csv files
kapture_linesep = '\n'


def get_version_from_header(header_string: str) -> Optional[str]:
    """
    Get the kapture format version from the given header string.

    :param header_string: a header string
    :return: version as string (i.e. '2.1') if found, None otherwise
    """
    m = re.search(KAPTURE_FORMAT_PARSING_RE, header_string)
    if m:
        return m['version']
    return None


def get_version_from_csv_file(csv_file_path: str) -> Optional[str]:
    """
    Get the kapture format version from a file.

    :param csv_file_path: path to the csv file
    :return: version as string (i.e. '2.1') if found, None otherwise
    """
    if path.isfile(csv_file_path):
        with open(csv_file_path) as f:
            first_line = f.readline()
            return get_version_from_header(first_line)
    return None


def current_format_version() -> Optional[str]:
    """
    Get the current format version

    :return: format version
    """
    return get_version_from_header(KAPTURE_FORMAT_1)


def kapture_format_version(kapture_dirpath: str) -> Optional[str]:
    """
    Reads kapture format version.

    :param kapture_dirpath: kapture directory root path
    :return: kapture format version if found.
    """
    sensors_file_path = path.join(kapture_dirpath, CSV_FILENAMES[kapture.Sensors])
    return get_version_from_csv_file(sensors_file_path)


def float_safe(representation) -> Optional[float]:
    """
    Safe float cast
    https://stackoverflow.com/questions/6330071/safe-casting-in-python

    :param representation: to cast
    :return: float value if the value was castable, or None
    """
    try:
        return float(representation)
    except (ValueError, TypeError):
        return None


def float_array_or_none(representation_list) -> Optional[List[float]]:
    """
    Safe cast of list of float representations
    https://stackoverflow.com/questions/6330071/safe-casting-in-python

    :param representation_list: list of values to convert
    :return: an array of floats or None if a single one is invalid
    """
    array = [float_safe(v) for v in representation_list]
    return array if not any(v is None for v in array) else None


def table_to_file(file, table, header=None, padding=None) -> int:
    """
    Writes the given table (list of list) into a file.
            The file must be previously open as write mode.
            If table is a generator, must be valid at runtime.

    :param file: file id opened in write mode (with open(filepath, 'w') as file:)
    :param table: an iterable of iterable
    :param header: row added at the beginning of the file (+\n)
    :param padding: the padding of each column as a list of int of same size of the rows of the table.
    :return: number of records written
    """
    if header:
        file.write(KAPTURE_FORMAT_1 + kapture_linesep)
        file.write(header + kapture_linesep)
    nb_records = 0
    for row in table:
        if padding:
            row = [str(v).rjust(padding[i]) for i, v in enumerate(row)]
        file.write(', '.join(f'{v}' for v in row) + kapture_linesep)
        nb_records += 1
    return nb_records


SPLIT_PATTERN = re.compile(r'\s*,\s*')


def table_from_file(file):
    """
    Returns an iterable of iterable (generator) on the opened file.
        Be aware that the returned generator is valid as long as file is valid.

    :param file: file id opened in read mode (with open(filepath, 'r') as file:)
    :return: an iterable of iterable on kapture objects values
    """
    table = file.readlines()
    if gettrace is not None and gettrace():
        table = (re.split(SPLIT_PATTERN, l1.rstrip("\n\r")) for l1 in table if l1.strip() and not l1.startswith('#'))
        return list(table)

    # remove comment lines, empty lines and trim trailing EOL
    table = (l1.rstrip("\n\r") for l1 in table if l1.strip() and not l1.startswith('#'))
    # then split comma (and trim afterwards spaces)
    content = []
    for line in table:
        words = line.split(',')
        words = list(word.strip() for word in words)
        content.append(words)
    return content


def get_last_line(opened_file: io.TextIOBase, max_line_size: int = 128) -> str:
    """
    Get the last line of an opened text file

    :param opened_file: an opened file (as returned by the builtin open)
    :param max_line_size: the maximum size of a line
    :return: last line if found, empty string otherwise
    """
    distance_to_end = 2 * max_line_size
    current_pos = opened_file.tell()
    # Check file size
    opened_file.seek(0, os.SEEK_END)
    file_size = opened_file.tell()
    if file_size > distance_to_end:
        # If we have a big file: skip towards the end
        opened_file.seek(file_size - distance_to_end, os.SEEK_SET)
    else:
        # back to start of file
        opened_file.seek(0, os.SEEK_SET)
    line = opened_file.readline()
    last_line = line
    while line:
        line = opened_file.readline()
        if line:
            last_line = line
    # back to position at the call of the function
    opened_file.seek(current_pos, os.SEEK_SET)
    return last_line


########################################################################################################################
# poses  ###############################################################################################################
def pose_to_list(pose: kapture.PoseTransform) -> List[Union[float, str]]:
    """
    Converts a 6D pose to a list of floats to save in CSV

    :param pose: 6D pose
    :return: list of float
    """
    assert (isinstance(pose, kapture.PoseTransform))
    rotation = pose.r_raw if pose.r is not None else 4 * ['']
    translation = pose.t_raw if pose.t is not None else 3 * ['']
    return rotation + translation


########################################################################################################################
# Sensor ###############################################################################################################
def sensor_to_list(sensor: kapture.Sensor) -> List[str]:
    """
    Converts a sensor into a list of strings to save in CSV

    :param sensor:
    :return: list of strings
    """
    assert (isinstance(sensor, kapture.Sensor))
    # sensor_id, name, model, [model_params]+
    fields = [sensor.name or ''] + [sensor.sensor_type] + [str(v) for v in sensor.sensor_params]
    return fields


########################################################################################################################
# Sensors ##############################################################################################################
def sensors_to_file(filepath: str, sensors: kapture.Sensors) -> None:
    """
    Writes the sensors to CSV file.

    :param filepath: input file path
    :param sensors: input sensors
    """
    assert (isinstance(sensors, kapture.Sensors))
    header = '# sensor_id, name, sensor_type, [sensor_params]+'
    table = ([sensor_id] + sensor_to_list(sensor)
             for sensor_id, sensor in sensors.items())

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header)


def sensors_from_file(filepath: str) -> kapture.Sensors:
    """
    Reads sensors from CSV file.

    :param filepath: input file path
    :return: sensors
    """
    sensors = kapture.Sensors()
    with open(filepath) as file:
        table = table_from_file(file)
        # sensor_id, name, sensor_type, [sensor_params]+'
        for sensor_id, name, sensor_type, *sensor_params in table:
            sensor = kapture.create_sensor(sensor_type=sensor_type, sensor_params=sensor_params, name=name)
            sensors[sensor_id] = sensor

    return sensors


########################################################################################################################
# Rig ##################################################################################################################
def rigs_to_file(filepath: str, rigs: kapture.Rigs) -> None:
    """
    Writes rigs to CSV file.

    :param filepath:
    :param rigs:
    """
    assert (isinstance(rigs, kapture.Rigs))
    header = '# rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz'
    padding = PADDINGS['device_id'] + PADDINGS['device_id'] + PADDINGS['pose']
    table = ([rig_id, sensor_id] + pose_to_list(pose)
             for rig_id, rig in rigs.items()
             for sensor_id, pose in rig.items())

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        table_to_file(file, table, header=header, padding=padding)


def rigs_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None) -> kapture.Rigs:
    """
    Reads rigs from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid sensor ids.
                        If a rig id collides one of them, raise error.
                        If a sensor in rig is not in sensor_ids, it is ignored.
    :return: rigs
    """
    # rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz
    rigs = kapture.Rigs()
    with open(filepath) as file:
        table = table_from_file(file)
        for rig_id, sensor_id, qw, qx, qy, qz, tx, ty, tz in table:
            if rig_id in sensor_ids:
                raise ValueError(f'collision between a sensor ID and rig ID ({rig_id})')
            rotation = float_array_or_none([qw, qx, qy, qz])
            translation = float_array_or_none([tx, ty, tz])
            pose = kapture.PoseTransform(rotation, translation)
            rigs[str(rig_id), sensor_id] = pose

    if sensor_ids is not None:
        # expunge all undesired sensors
        rig_ids = set(rigs)
        for rig_id in rig_ids:
            for sensor_id in set(rigs[rig_id]):
                if sensor_id not in sensor_ids and sensor_id not in rig_ids:
                    logger.debug(f'dropping sensor {sensor_id} from rig {rig_id} because it is unknown sensor.')
                    del rigs[rig_id][sensor_id]

    return rigs


########################################################################################################################
# Trajectories #########################################################################################################
def trajectories_to_file(filepath: str, trajectories: kapture.Trajectories) -> None:
    """
    Writes trajectories to CSV file.

    :param filepath:
    :param trajectories:
    """
    assert (isinstance(trajectories, kapture.Trajectories))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, qw, qx, qy, qz, tx, ty, tz'
    padding = PADDINGS['timestamp'] + PADDINGS['device_id'] + PADDINGS['pose']
    table = (
        [timestamp, sensor_id] + pose_to_list(trajectories[(timestamp, sensor_id)])
        for timestamp, sensor_id in sorted(trajectories.key_pairs())
    )

    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header, padding=padding)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(trajectories)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def trajectories_from_file(filepath: str, device_ids: Optional[Set[str]] = None) -> kapture.Trajectories:
    """
    Reads trajectories from CSV file.

    :param filepath: input file path
    :param device_ids: input set of valid device ids (rig or sensor).
                        If the trajectories contains unknown devices, they will be ignored.
                        If no device_ids given, everything is loaded.
    :return: trajectories
    """
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        trajectories = kapture.Trajectories()
        # timestamp, device_id, qw, qx, qy, qz, tx, ty, tz
        for timestamp, device_id, qw, qx, qy, qz, tx, ty, tz in table:
            if device_ids is not None and device_id not in device_ids:
                # just ignore
                continue
            pose = kapture.PoseTransform.__new__(kapture.PoseTransform)
            if qw != '' and qx != '' and qy != '' and qz != '':
                rotation = quaternion.from_float_array([float(qw), float(qx), float(qy), float(qz)])
            else:
                rotation = None
            pose._r = rotation

            if tx != '' and ty != '' and tz != '':
                trans = np.array([[float(tx)], [float(ty)], [float(tz)]], dtype=np.float)
            else:
                trans = None
            pose._t = trans
            trajectories.setdefault(int(timestamp), {})[device_id] = pose
            nb_records += 1
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.Trajectories} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return trajectories


########################################################################################################################
# Records Camera #######################################################################################################
def records_camera_to_file(filepath: str, records_camera: kapture.RecordsCamera) -> None:
    """
    Writes records_camera to CSV file.

    :param filepath:
    :param records_camera:
    """
    assert (isinstance(records_camera, kapture.RecordsCamera))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, image_path'
    table = (
        [timestamp, sensor_id] + [records_camera[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_camera.key_pairs())
    )
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records_camera)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def records_camera_from_file(filepath: str, camera_ids: Optional[Set[str]] = None) -> kapture.RecordsCamera:
    """
    Reads records_camera from CSV file.

    :param filepath: input file path
    :param camera_ids: input set of valid camera device ids.
                        If the records_camera contains unknown devices, they will be ignored.
                        If not given, all cameras are loaded.
    :return: camera records
    """
    records_camera = kapture.RecordsCamera()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # timestamp, device_id, image_path
        for timestamp, device_id, image_path in table:
            if camera_ids is not None and device_id not in camera_ids:
                # just ignore
                continue
            records_camera[(int(timestamp), str(device_id))] = image_path
            nb_records += 1
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.RecordsCamera} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records_camera


########################################################################################################################
# Records Depth #######################################################################################################
def records_depth_to_file(filepath: str, records_depth: kapture.RecordsDepth) -> None:
    """
    Writes records_depth to CSV file.

    :param filepath:
    :param records_depth:
    """
    assert (isinstance(records_depth, kapture.RecordsDepth))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, depth_map_path'
    table = (
        [timestamp, sensor_id] + [records_depth[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_depth.key_pairs())
    )
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records_depth)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def records_depth_from_file(filepath: str, camera_ids: Optional[Set[str]] = None) -> kapture.RecordsDepth:
    """
    Reads records_depth from CSV file.

    :param filepath: input file path
    :param camera_ids: input set of valid camera device ids.
                        If the records_camera contains unknown devices, they will be ignored.
                        If not given, all cameras are loaded.
    :return: camera records
    """
    records_depth = kapture.RecordsDepth()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # timestamp, device_id, image_path
        for timestamp, device_id, depth_map_path in table:
            if camera_ids is not None and device_id not in camera_ids:
                # just ignore
                continue
            records_depth[(int(timestamp), str(device_id))] = depth_map_path
            nb_records += 1
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.RecordsDepth} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records_depth


########################################################################################################################
# Records Lidar ########################################################################################################
def records_lidar_to_file(filepath: str, records_lidar: kapture.RecordsLidar) -> None:
    """
    Writes records_lidar to CSV file.

    :param filepath:
    :param records_lidar:
    """
    assert (isinstance(records_lidar, kapture.RecordsLidar))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, point_cloud_path'
    table = (
        [timestamp, sensor_id] + [records_lidar[(timestamp, sensor_id)]]
        for timestamp, sensor_id in sorted(records_lidar.key_pairs())
    )
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records_lidar)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def records_lidar_from_file(filepath: str, lidar_ids: Optional[Set[str]] = None
                            ) -> kapture.RecordsLidar:
    """
    Reads records_lidar from CSV file.

    :param filepath: input file path
    :param lidar_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                            If omitted, then it loads all devices.
    :return: Lidar records
    """
    records_lidar = kapture.RecordsLidar()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # timestamp, device_id, point_cloud_path
        for timestamp, device_id, point_cloud_path in table:
            if lidar_ids is not None and device_id not in lidar_ids:
                # just ignore
                continue
            records_lidar[(int(timestamp), str(device_id))] = point_cloud_path
            nb_records += 1
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.RecordsLidar} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records_lidar


########################################################################################################################
def records_generic_to_file(filepath: str, records: kapture.RecordsBase) -> None:
    """
        Writes records_wifi to file

        :param filepath: path where to save records file.
        :param records:
    """
    assert (isinstance(records, kapture.RecordsBase))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, ' + ', '.join(f.name for f in records.record_type.fields())
    table = []
    for timestamp, sensor_id, record in kapture.flatten(records, is_sorted=True):
        table.append([timestamp, sensor_id] + [str(v) for v in record.astuple()])
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def records_generic_from_file(records_type: Type, filepath: str, sensor_ids: Optional[Set[str]] = None
                              ) -> Union[kapture.RecordsBase,
                                         kapture.RecordsGnss,
                                         kapture.RecordsGyroscope,
                                         kapture.RecordsAccelerometer,
                                         kapture.RecordsMagnetic]:
    """
    Reads Records data from CSV file.

    :param records_type: type of records expected (eg RecordsWifi)
    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                     If omitted, then it loads all devices.
    :return: records
    """
    records = records_type()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        # timestamp, device_id, *
        nb_records = 0
        for timestamp, device_id, *data in table:
            timestamp = int(timestamp)
            device_id = str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            records[timestamp, device_id] = records_type.record_type(*data)
            nb_records += 1

    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {records_type} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records


# Records Wifi #########################################################################################################
def records_wifi_to_file(filepath: str, records_wifi: kapture.RecordsWifi) -> None:
    """
    Writes records_wifi to file

    :param filepath:
    :param records_wifi:
    """
    assert (isinstance(records_wifi, kapture.RecordsWifi))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end'
    table = []
    for timestamp, sensor_id in sorted(records_wifi.key_pairs()):
        for bssid, record in records_wifi[timestamp, sensor_id].items():
            table.append([timestamp, sensor_id, bssid] + [str(v) for v in record.astuple()])
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records_wifi)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def records_wifi_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                           ) -> kapture.RecordsWifi:
    """
    Reads RecordsWifi from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                            If omitted, then it loads all devices.
    :return: Wifi records
    """
    records_wifi = kapture.RecordsWifi()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end
        for timestamp, device_id, BSSID, frequency, RSSI, SSID, scan_time_start, scan_time_end in table:
            timestamp, device_id = int(timestamp), str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            if (timestamp, device_id) not in records_wifi:
                records_wifi[timestamp, device_id] = kapture.RecordWifi()
            records_wifi[timestamp, device_id][BSSID] = kapture.RecordWifiSignal(
                frequency, RSSI, SSID, scan_time_start, scan_time_end)
            nb_records += 1

    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.RecordsWifi} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records_wifi


# Records Bluetooth ####################################################################################################
def records_bluetooth_to_file(filepath: str, records_bluetooth: kapture.RecordsBluetooth) -> None:
    """
    Writes Bluetooth records to file

    :param filepath: output file path.
    :param records_bluetooth: records to save
    """
    assert (isinstance(records_bluetooth, kapture.RecordsBluetooth))
    saving_start = datetime.datetime.now()
    header = '# timestamp, device_id, address, RSSI, name'
    table = []
    for timestamp, sensor_id in sorted(records_bluetooth.key_pairs()):
        for address, bt_record in records_bluetooth[timestamp, sensor_id].items():
            table.append([timestamp, sensor_id, address] + [str(v) for v in bt_record.astuple()])
    with open(filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(records_bluetooth)}'
                     f' in {saving_elapsed.total_seconds():.3f} seconds'.replace(',', ' '))


def records_bluetooth_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                                ) -> kapture.RecordsBluetooth:
    """
    Reads Bluetooth records from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: Bluetooth records
    """
    records_bluetooth = kapture.RecordsBluetooth()
    loading_start = datetime.datetime.now()
    with open(filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # timestamp, device_id, address, RSSI, name
        for timestamp, device_id, address, RSSI, name in table:
            timestamp, device_id = int(timestamp), str(device_id)
            if sensor_ids is not None and device_id not in sensor_ids:
                # just ignore
                continue
            if (timestamp, device_id) not in records_bluetooth:
                records_bluetooth[timestamp, device_id] = kapture.RecordBluetooth()
            records_bluetooth[timestamp, device_id][address] = kapture.RecordBluetoothSignal(rssi=RSSI, name=name)
            nb_records += 1

    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.RecordsBluetooth} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return records_bluetooth


# GNSS #################################################################################################################
def records_gnss_to_file(filepath: str, records_gnss: kapture.RecordsGnss) -> None:
    """
    Writes Gnss records to file

    :param filepath: output file path.
    :param records_gnss: records to save
    """
    records_generic_to_file(filepath, records_gnss)


def records_gnss_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                           ) -> kapture.RecordsGnss:
    """
    Reads Gnss records from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: Gnss records
    """
    return records_generic_from_file(kapture.RecordsGnss, filepath, sensor_ids)


# Accelerometer ########################################################################################################
def records_accelerometer_to_file(filepath: str, records_accelerometer: kapture.RecordsAccelerometer) -> None:
    """
    Writes accelerometer records to file

    :param filepath: output file path.
    :param records_accelerometer: records to save
    """
    records_generic_to_file(filepath, records_accelerometer)


def records_accelerometer_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                                    ) -> kapture.RecordsAccelerometer:
    """
    Reads accelerometer records from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: accelerometer records
    """
    return records_generic_from_file(kapture.RecordsAccelerometer, filepath, sensor_ids)


# Gyroscope ########################################################################################################
def records_gyroscope_to_file(filepath: str, records_gyroscope: kapture.RecordsGyroscope) -> None:
    """
    Writes gyroscope records to file

    :param filepath: output file path.
    :param records_gyroscope: records to save
    """
    records_generic_to_file(filepath, records_gyroscope)


def records_gyroscope_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                                ) -> kapture.RecordsGyroscope:
    """
    Reads gyroscope records from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: gyroscope records
    """
    return records_generic_from_file(kapture.RecordsGyroscope, filepath, sensor_ids)


# Magnetic ########################################################################################################
def records_magnetic_to_file(filepath: str, records_magnetic: kapture.RecordsMagnetic) -> None:
    """
    Writes magnetic records to file

    :param filepath: output file path.
    :param records_magnetic: records to save
    """
    records_generic_to_file(filepath, records_magnetic)


def records_magnetic_from_file(filepath: str, sensor_ids: Optional[Set[str]] = None
                               ) -> kapture.RecordsMagnetic:
    """
    Reads magnetic records from CSV file.

    :param filepath: input file path
    :param sensor_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
                        If omitted, then it loads all devices.
    :return: magnetic records
    """
    return records_generic_from_file(kapture.RecordsMagnetic, filepath, sensor_ids)


########################################################################################################################
# features #############################################################################################################
def image_features_set_from_dir(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures]],
        feature_type: str,
        kapture_dirpath: str,
        image_filenames: Optional[Set[str]]
) -> Set[str]:
    """
    Reads and builds ImageFeatures from images_filenames if given, or directly from actual files in  root_dirpath.

    :param kapture_type: kapture class type.
    :param feature_type: name of the feature
    :param kapture_dirpath: input path to kapture root directory.
    :param image_filenames: None or set of image relative paths.
    :return: Set
    """

    if image_filenames is None:
        # images_path is empty, so populates all feature files
        image_filenames_generator = kapture.io.features.image_ids_from_feature_dirpath(kapture_type,
                                                                                       feature_type,
                                                                                       kapture_dirpath)
    else:
        # filter only existing files
        image_filenames_generator = (
            image_name
            for image_name in image_filenames
            if path.exists(kapture.io.features.get_features_fullpath(kapture_type,
                                                                     feature_type,
                                                                     kapture_dirpath,
                                                                     image_name)))
    return set(image_filenames_generator)


def image_features_set_from_tar(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures]],
        tar_handler: TarHandler,
        image_filenames: Optional[Set[str]]
) -> Set[str]:
    """
    Reads and builds ImageFeatures from images_filenames if given, or directly from actual files in the tar.

    :param kapture_type: kapture class type.
    :param tar_handler: opened tar reference
    :param image_filenames: None or set of image relative paths.
    :return: Set
    """
    image_filenames_generator = kapture.io.features.image_ids_from_feature_tar(kapture_type, tar_handler)
    if image_filenames is not None:
        image_filenames_generator = (image_name
                                     for image_name in image_filenames_generator
                                     if image_name in image_filenames)
    return set(image_filenames_generator)


########################################################################################################################
# keypoints ############################################################################################################
KeypointsConfig = namedtuple('KeypointsConfig', ['name', 'dtype', 'dsize'])


def keypoints_to_file(config_filepath: str, keypoints: kapture.Keypoints) -> None:
    """
    Writes keypoints to CSV file.

    :param config_filepath:
    :param keypoints:
    """
    os.makedirs(path.dirname(config_filepath), exist_ok=True)
    header = "# name, dtype, dsize"
    dtype = str(keypoints.dtype) if isinstance(keypoints.dtype, np.dtype) else keypoints.dtype.__name__
    line = [keypoints.type_name, dtype, str(keypoints.dsize)]
    with open(config_filepath, 'wt') as file:
        table_to_file(file, [line], header=header)


def keypoints_config_from_file(config_filepath: str) -> KeypointsConfig:
    """
    Reads keypoints config files, ie. name, data type, and data size.
    :param config_filepath: input path to config file to read.
    :return: config The read config.
    """
    if not path.exists(config_filepath):
        raise FileNotFoundError(f'{path.basename(config_filepath)} file is missing')

    with open(config_filepath, 'rt') as file:
        table = table_from_file(file)
        line = list(table)[0]
        assert len(line) == 3
        name, dtype, dsize = line[0], line[1], int(line[2])

    # try to list all possible type from numpy that can be used in eval(dtype)
    from numpy import float, float32, float64, int32, uint8  # noqa: F401
    if isinstance(type(eval(dtype)), type):
        dtype = eval(dtype)
    else:
        raise ValueError('Expect data type ')
    return KeypointsConfig(name, dtype, dsize)


def keypoints_from_dir(keypoints_type: str,
                       kapture_dirpath: str,
                       images_paths: Optional[Set[str]],
                       tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> kapture.Keypoints:
    """
    Reads and builds keypoints from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :param tar_handler: collection of preloaded tar archives
    :return: Keypoints
    """
    # make config_filepath from data_dirpath
    config_filepath = get_feature_csv_fullpath(kapture.Keypoints, keypoints_type, kapture_dirpath)
    config = keypoints_config_from_file(config_filepath)

    # exist as tar ?
    tar_local_handler = retrieve_tar_handler_from_collection(kapture.Keypoints, keypoints_type, tar_handler)

    if tar_local_handler is not None:
        image_filenames = image_features_set_from_tar(kapture.Keypoints, tar_local_handler, images_paths)
    else:
        # exist tar ? -> fire warning
        tar_path = get_feature_tar_fullpath(kapture.Keypoints, keypoints_type, kapture_dirpath)
        if os.path.isfile(tar_path):
            getLogger().warning(f'{tar_path} exist but no handler was given so it is ignored and loaded from dir')
        image_filenames = image_features_set_from_dir(kapture.Keypoints,
                                                      keypoints_type,
                                                      kapture_dirpath,
                                                      images_paths)
    return kapture.Keypoints(config.name, config.dtype, config.dsize, image_filenames)


########################################################################################################################
# descriptors ##########################################################################################################
DescriptorsConfig = namedtuple('DescriptorsConfig', ['name', 'dtype', 'dsize', 'keypoints_type', 'metric_type'])


def descriptors_to_file(config_filepath: str, descriptors: kapture.Descriptors) -> None:
    """
    Writes descriptors to CSV file.
    :param config_filepath:
    :param descriptors:
    """
    os.makedirs(path.dirname(config_filepath), exist_ok=True)
    header = "# name, dtype, dsize, keypoints_type, metric_type"
    dtype = str(descriptors.dtype) if isinstance(descriptors.dtype, np.dtype) else descriptors.dtype.__name__
    line = [descriptors.type_name, dtype, str(descriptors.dsize), descriptors.keypoints_type, descriptors.metric_type]
    with open(config_filepath, 'wt') as file:
        table_to_file(file, [line], header=header)


def descriptors_config_from_file(config_filepath: str) -> DescriptorsConfig:
    """
    Reads descriptors config files.
    :param config_filepath: input path to config file to read.
    :return: config The read config.
    """
    if not path.exists(config_filepath):
        raise FileNotFoundError(f'{path.basename(config_filepath)} file is missing')

    with open(config_filepath, 'rt') as file:
        table = table_from_file(file)
        line = list(table)[0]
        assert len(line) == 5
        name, dtype, dsize, keypoints_type, metric_type = line[0], line[1], int(line[2]), line[3], line[4]

    # try to list all possible type from numpy that can be used in eval(dtype)
    from numpy import float, float32, float64, int32, uint8  # noqa: F401
    if isinstance(type(eval(dtype)), type):
        dtype = eval(dtype)
    else:
        raise ValueError('Expect data type ')
    return DescriptorsConfig(name, dtype, dsize, keypoints_type, metric_type)


def descriptors_from_dir(descriptors_type: str, kapture_dirpath: str, images_paths: Optional[Set[str]],
                         tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> kapture.Descriptors:
    """
    Reads and builds descriptors from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param descriptors_type: type of descriptors to export, name of the descriptors subfolder
    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :param tar_handler: collection of preloaded tar archives
    :return: Descriptors
    """
    # make config_filepath from data_dirpath
    config_filepath = get_feature_csv_fullpath(kapture.Descriptors, descriptors_type, kapture_dirpath)
    config = descriptors_config_from_file(config_filepath)

    # exist as tar ?
    tar_local_handler = retrieve_tar_handler_from_collection(kapture.Descriptors, descriptors_type, tar_handler)

    if tar_local_handler is not None:
        image_filenames = image_features_set_from_tar(kapture.Descriptors, tar_local_handler, images_paths)
    else:
        # exist tar ? -> fire warning
        tar_path = get_feature_tar_fullpath(kapture.Descriptors, descriptors_type, kapture_dirpath)
        if os.path.isfile(tar_path):
            getLogger().warning(f'{tar_path} exist but no handler was given so it is ignored and loaded from dir')
        image_filenames = image_features_set_from_dir(kapture.Descriptors,
                                                      descriptors_type,
                                                      kapture_dirpath,
                                                      images_paths)
    return kapture.Descriptors(config.name,
                               config.dtype,
                               config.dsize,
                               config.keypoints_type,
                               config.metric_type,
                               image_filenames)


########################################################################################################################
# global_features ######################################################################################################
GlobalFeaturesConfig = namedtuple('GlobalFeaturesConfig', ['name', 'dtype', 'dsize', 'metric_type'])


def global_features_to_file(config_filepath: str, global_features: kapture.GlobalFeatures) -> None:
    """
    Writes global features to CSV file.

    :param config_filepath:
    :param global_features:
    """
    os.makedirs(path.dirname(config_filepath), exist_ok=True)
    header = "# name, dtype, dsize, metric_type"
    if isinstance(global_features.dtype, np.dtype):
        dtype = str(global_features.dtype)
    else:
        dtype = global_features.dtype.__name__
    line = [global_features.type_name, dtype, str(global_features.dsize), global_features.metric_type]
    with open(config_filepath, 'wt') as file:
        table_to_file(file, [line], header=header)


def global_features_config_from_file(config_filepath: str) -> GlobalFeaturesConfig:
    """
    Reads global_features config files.
    :param config_filepath: input path to config file to read.
    :return: config The read config.
    """
    if not path.exists(config_filepath):
        raise FileNotFoundError(f'{path.basename(config_filepath)} file is missing')

    with open(config_filepath, 'rt') as file:
        table = table_from_file(file)
        line = list(table)[0]
        assert len(line) == 4
        name, dtype, dsize, metric_type = line[0], line[1], int(line[2]), line[3]

    # try to list all possible type from numpy that can be used in eval(dtype)
    from numpy import float, float32, float64, int32, uint8  # noqa: F401
    if isinstance(type(eval(dtype)), type):
        dtype = eval(dtype)
    else:
        raise ValueError('Expect data type ')
    return GlobalFeaturesConfig(name, dtype, dsize, metric_type)


def global_features_from_dir(global_features_type: str,
                             kapture_dirpath: str,
                             images_paths: Set[str],
                             tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> kapture.GlobalFeatures:
    """
    Reads and builds Global features from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param global_features_type: type of global_features, name of the global_features subfolder
    :param kapture_dirpath: root path of kapture
    :param images_paths: optional list of image file names
    :param tar_handler: collection of preloaded tar archives
    :return: Global features
    """
    # make config_filepath from data_dirpath
    config_filepath = get_feature_csv_fullpath(kapture.GlobalFeatures, global_features_type, kapture_dirpath)
    config = global_features_config_from_file(config_filepath)

    # exist as tar ?
    tar_local_handler = retrieve_tar_handler_from_collection(kapture.GlobalFeatures, global_features_type, tar_handler)

    if tar_local_handler is not None:
        image_filenames = image_features_set_from_tar(kapture.GlobalFeatures, tar_local_handler, images_paths)
    else:
        # exist tar ? -> fire warning
        tar_path = get_feature_tar_fullpath(kapture.GlobalFeatures, global_features_type, kapture_dirpath)
        if os.path.isfile(tar_path):
            getLogger().warning(f'{tar_path} exist but no handler was given so it is ignored and loaded from dir')
        image_filenames = image_features_set_from_dir(kapture.GlobalFeatures,
                                                      global_features_type,
                                                      kapture_dirpath,
                                                      images_paths)
    return kapture.GlobalFeatures(config.name,
                                  config.dtype,
                                  config.dsize,
                                  config.metric_type,
                                  image_filenames)


########################################################################################################################
# matches ##############################################################################################################
def matches_from_dir(
        keypoints_type: str,
        kapture_dirpath: str,
        image_filenames: Optional[Set[str]] = None,
        matches_pairsfile_path: Optional[str] = None,
        tar_handler: Optional[Union[TarCollection, TarHandler]] = None) -> kapture.Matches:
    """
    Reads and builds Matches from images_filenames if given, or directly from actual files in kapture_dirpath.

    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param kapture_dirpath: root path of kapture
    :param image_filenames: optional list of image file names
    :param matches_pairsfile_path: text file in the csv format; where each line is image_name1, image_name2, score
    :param tar_handler: collection of preloaded tar archives
    :return: Matches
    """
    # exist as tar ?
    tar_local_handler = retrieve_tar_handler_from_collection(kapture.Matches, keypoints_type, tar_handler)
    loading_start = datetime.datetime.now()
    if tar_local_handler is not None:
        match_pairs_generator = kapture.io.features.matching_pairs_from_tar(tar_local_handler)
        if matches_pairsfile_path is not None:
            with open(matches_pairsfile_path, 'r') as fid:
                table = table_from_file(fid)
                valid_pairs = set((query_name, map_name) if query_name < map_name else (map_name, query_name)
                                  for query_name, map_name, _ in table)
                match_pairs_generator = (image_pair
                                         for image_pair in match_pairs_generator
                                         if image_pair in valid_pairs)
    else:
        # exist tar ? -> fire warning
        tar_path = get_feature_tar_fullpath(kapture.Matches, keypoints_type, kapture_dirpath)
        if os.path.isfile(tar_path):
            getLogger().warning(f'{tar_path} exist but no handler was given so it is ignored and loaded from dir')
        if matches_pairsfile_path is None:
            # populate files from disk
            match_pairs_generator = kapture.io.features.matching_pairs_from_dirpath(keypoints_type, kapture_dirpath)
        else:
            with open(matches_pairsfile_path, 'r') as fid:
                table = table_from_file(fid)
                # get matches list from pairsfile
                match_pairs_generator = ((query_name, map_name) if query_name < map_name else (map_name, query_name)
                                         for query_name, map_name, _ in table)
                # keeps only the one that actually exists on disk
                match_pairs_generator = (image_pair
                                         for image_pair in match_pairs_generator
                                         if path.isfile(kapture.io.features.get_matches_fullpath(image_pair,
                                                                                                 keypoints_type,
                                                                                                 kapture_dirpath))
                                         )
    if image_filenames is not None:
        # retains only files that correspond to known images
        match_pairs_generator = (
            image_pair
            for image_pair in match_pairs_generator
            if image_pair[0] in image_filenames and image_pair[1] in image_filenames
        )
    match_pairs = set(match_pairs_generator)
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{len(match_pairs)} {kapture.Matches} in {loading_elapsed.total_seconds():.3f} seconds')
    return kapture.Matches(match_pairs)


########################################################################################################################
# points3d #############################################################################################################
def points3d_to_file(filepath: str, points3d: kapture.Points3d) -> None:
    """
    Writes 3d points to CSV file.

    :param filepath: path to CSV file
    :param points3d: the 3d points
    """
    assert isinstance(points3d, kapture.Points3d)
    os.makedirs(path.dirname(filepath), exist_ok=True)
    header = KAPTURE_FORMAT_1[2:] + kapture_linesep + 'X, Y, Z, R, G, B'
    np.savetxt(filepath, points3d.as_array(), delimiter=',', header=header)


def points3d_from_file(filepath: str) -> kapture.Points3d:
    """
    Reads 3d points from CSV file.

    :param filepath: path to CSV file
    :return: the 3d points
    """

    loading_start = datetime.datetime.now()
    data = np.loadtxt(filepath, dtype=np.float, delimiter=',', comments='#')
    data = data.reshape((-1, 6))  # make sure of the shape, even if single line file.
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{kapture.Points3d} in {loading_elapsed.total_seconds():.3f} seconds')
    return kapture.Points3d(data)


########################################################################################################################
# observations #########################################################################################################
def observations_to_file(observations_filepath: str, observations: kapture.Observations) -> None:
    """
    Writes observations to CSV file.

    :param observations_filepath: input path to CSV file of observation to write.
                                    Containing directory is created if needed.
    :param observations: input observations to be written.
    """
    assert path.basename(observations_filepath) == 'observations.txt'
    assert isinstance(observations, kapture.Observations)
    saving_start = datetime.datetime.now()
    header = '# point3d_id, keypoints_type, [image_path, feature_id]*'
    table = (
        [str(point3d_idx), str(keypoints_type)] + [str(k)
                                                   for pair in observations[point3d_idx, keypoints_type]
                                                   for k in pair]
        for point3d_idx, keypoints_type in sorted(observations.key_pairs(), key=lambda x: (x[0], x[1]))
    )
    os.makedirs(path.dirname(observations_filepath), exist_ok=True)
    with open(observations_filepath, 'w') as file:
        nb_records = table_to_file(file, table, header=header)
        saving_elapsed = datetime.datetime.now() - saving_start
        logger.debug(f'wrote {nb_records:12,d} {type(observations)} in {saving_elapsed.total_seconds():.3f} seconds'
                     .replace(',', ' '))


def observations_from_file(observations_filepath: str, loaded_keypoints: Optional[Dict[str, Set[str]]] = None)\
        -> kapture.Observations:
    """
    Reads observations from CSV file.

    :param observations_filepath: path to CSV file to read.
    :param loaded_keypoints: input set of image names (ids) that have keypoints.
                                        If given, used to filter out irrelevant observations.
                                        You can get from set(kapture.keypoints)
    :return: observations
    """
    assert path.basename(observations_filepath) == 'observations.txt'
    assert loaded_keypoints is None \
        or (isinstance(loaded_keypoints, dict) and len(loaded_keypoints) > 0)
    assert loaded_keypoints is None or all([isinstance(keypoints, set) for keypoints in loaded_keypoints.values()])

    observations = kapture.Observations()
    loading_start = datetime.datetime.now()
    with open(observations_filepath) as file:
        table = table_from_file(file)
        nb_records = 0
        # point3d_id, keypoints_type, [image_path, feature_id]*
        for points3d_id_str, keypoints_type, *pairs in table:
            if loaded_keypoints is not None and \
                    (keypoints_type not in loaded_keypoints or len(loaded_keypoints[keypoints_type]) == 0):
                continue
            points3d_id = int(points3d_id_str)
            if len(pairs) > 1:
                image_paths = pairs[0::2]
                keypoints_ids = pairs[1::2]
                for image_path, keypoint_id in zip(image_paths, keypoints_ids):
                    if loaded_keypoints is not None and image_path not in loaded_keypoints[keypoints_type]:
                        # image_path does not exist in kapture (perhaps it was removed), ignore it
                        continue
                    observations.add(points3d_id, keypoints_type, image_path, int(keypoint_id))
                    nb_records += 1
    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'{nb_records:12,d} {kapture.Observations} in {loading_elapsed.total_seconds():.3f} seconds'
                 .replace(',', ' '))
    return observations


########################################################################################################################
# Kapture Write ########################################################################################################
KAPTURE_ATTRIBUTE_WRITERS = {
    kapture.Sensors: sensors_to_file,
    kapture.Rigs: rigs_to_file,
    kapture.Trajectories: trajectories_to_file,
    kapture.RecordsCamera: records_camera_to_file,
    kapture.RecordsDepth: records_depth_to_file,
    kapture.RecordsLidar: records_lidar_to_file,
    kapture.RecordsWifi: records_wifi_to_file,
    kapture.RecordsBluetooth: records_bluetooth_to_file,
    kapture.RecordsGnss: records_gnss_to_file,
    kapture.RecordsAccelerometer: records_accelerometer_to_file,
    kapture.RecordsGyroscope: records_gyroscope_to_file,
    kapture.RecordsMagnetic: records_magnetic_to_file,
    kapture.Keypoints: keypoints_to_file,
    kapture.Descriptors: descriptors_to_file,
    kapture.GlobalFeatures: global_features_to_file,
    kapture.Points3d: points3d_to_file,
    kapture.Observations: observations_to_file,
}

KAPTURE_ATTRIBUTE_NAMES = {  # used to list attributes to be saved
    kapture.Sensors: 'sensors',
    kapture.Rigs: 'rigs',
    kapture.Trajectories: 'trajectories',
    kapture.RecordsCamera: 'records_camera',
    kapture.RecordsDepth: 'records_depth',
    kapture.RecordsLidar: 'records_lidar',
    kapture.RecordsWifi: 'records_wifi',
    kapture.RecordsBluetooth: 'records_bluetooth',
    kapture.RecordsGnss: 'records_gnss',
    kapture.RecordsAccelerometer: 'records_accelerometer',
    kapture.RecordsGyroscope: 'records_gyroscope',
    kapture.RecordsMagnetic: 'records_magnetic',
    kapture.Keypoints: 'keypoints',
    kapture.Descriptors: 'descriptors',
    kapture.GlobalFeatures: 'global_features',
    kapture.Points3d: 'points3d',
    kapture.Observations: 'observations',
}


def kapture_to_dir(kapture_dirpath: str, kapture_data: kapture.Kapture) -> None:
    """
    Saves kapture data to given directory.

    :param kapture_dirpath: kapture directory root path
    :param kapture_data: input kapture data
    """
    kapture_subtype_to_filepaths = {kapture_class: path.join(kapture_dirpath, filename)
                                    for kapture_class, filename in CSV_FILENAMES.items()}
    saving_start = datetime.datetime.now()
    # save each member of kapture data
    for kapture_class, kapture_member_name in KAPTURE_ATTRIBUTE_NAMES.items():
        part_data = kapture_data.__getattribute__(kapture_member_name)
        if part_data is not None and kapture_class in kapture_subtype_to_filepaths:
            # save it
            logger.debug(f'saving {kapture_member_name} ...')
            write_function = KAPTURE_ATTRIBUTE_WRITERS[kapture_class]
            write_function(kapture_subtype_to_filepaths[kapture_class], part_data)
        elif part_data is not None and kapture_class in FEATURES_CSV_FILENAMES:
            write_function = KAPTURE_ATTRIBUTE_WRITERS[kapture_class]
            for feature_type, features in part_data.items():
                # save it
                logger.debug(f'saving {kapture_member_name} : {feature_type} ...')
                filepath = path.join(kapture_dirpath, FEATURES_CSV_FILENAMES[kapture_class](feature_type))
                write_function(filepath, features)
    saving_elapsed = datetime.datetime.now() - saving_start
    logger.info(f'Saved in {saving_elapsed.total_seconds():.3f} seconds in "{kapture_dirpath}"')


# Kapture Read #########################################################################################################
# list all data members of kapture.
KAPTURE_LOADABLE_TYPES = {
    kapture.Sensors,
    kapture.Rigs,
    kapture.Trajectories,
    kapture.RecordsCamera,
    kapture.RecordsDepth,
    kapture.RecordsLidar,
    kapture.RecordsWifi,
    kapture.RecordsBluetooth,
    kapture.RecordsGnss,
    kapture.RecordsAccelerometer,
    kapture.RecordsGyroscope,
    kapture.RecordsMagnetic,
    kapture.Keypoints,
    kapture.Descriptors,
    kapture.GlobalFeatures,
    kapture.Matches,
    kapture.Points3d,
    kapture.Observations,
}


def kapture_from_dir(
        kapture_dir_path: str,
        matches_pairs_file_path: Optional[str] = None,
        skip_list: List[Type[Union[kapture.Rigs,
                                   kapture.Trajectories,
                                   kapture.RecordsCamera,
                                   kapture.RecordsDepth,
                                   kapture.RecordsLidar,
                                   kapture.RecordsWifi,
                                   kapture.RecordsBluetooth,
                                   kapture.RecordsGnss,
                                   kapture.RecordsAccelerometer,
                                   kapture.RecordsGyroscope,
                                   kapture.RecordsMagnetic,
                                   kapture.Keypoints,
                                   kapture.Descriptors,
                                   kapture.GlobalFeatures,
                                   kapture.Matches,
                                   kapture.Points3d,
                                   kapture.Observations
                                   ]]] = [],
        tar_handlers: Optional[TarCollection] = None
) -> kapture.Kapture:
    """
    Reads and return kapture data from directory.

    :param kapture_dir_path: kapture directory root path
    :param matches_pairs_file_path: text file in the csv format; where each line is image_name1, image_name2, score
    :param skip_list: Input option for expert only. Skip the load of specified parts.
    :param tar_handlers: collection of preloaded tar archives
    :return: kapture data read
    """
    if not path.isdir(kapture_dir_path):
        raise FileNotFoundError(f'No kapture directory {kapture_dir_path}')
    csv_file_paths = {dtype: path.join(kapture_dir_path, filename)
                      for dtype, filename in CSV_FILENAMES.items()}
    data_dir_paths = {dtype: path.join(kapture_dir_path, dir_name)
                      for dtype, dir_name in kapture.io.features.FEATURES_DATA_DIRNAMES.items()}

    # keep only those in load_only and that exists
    kapture_data_paths = {**data_dir_paths, **csv_file_paths}  # make sure files take precedence over dirs
    kapture_loadable_data = {
        kapture_type
        for kapture_type in KAPTURE_LOADABLE_TYPES
        if kapture_type not in skip_list and path.exists(kapture_data_paths[kapture_type])
    }

    kapture_data = kapture.Kapture()
    loading_start = datetime.datetime.now()
    # sensors
    sensor_ids = None
    sensors_file_path = csv_file_paths[kapture.Sensors]
    if sensors_file_path:
        logger.debug(f'loading sensors {sensors_file_path} ...')
        kapture_data.__version__ = get_version_from_csv_file(sensors_file_path)
        assert kapture_data.__version__ == current_format_version()
        kapture_data.sensors = sensors_from_file(sensors_file_path)
        sensor_ids = set(kapture_data.sensors.keys()) if kapture_data.sensors is not None else set()

    if sensor_ids is None:
        # no need to continue, everything else depends on sensors
        raise FileNotFoundError(f'File {sensors_file_path} is missing or empty in {kapture_dir_path}')

    # rigs
    if kapture.Rigs in kapture_loadable_data:
        rigs_file_path = csv_file_paths[kapture.Rigs]
        logger.debug(f'loading rigs {rigs_file_path} ...')
        assert sensor_ids is not None
        kapture_data.rigs = rigs_from_file(rigs_file_path, sensor_ids)
        # update sensor_ids with rig_id
        sensor_ids.update(kapture_data.rigs.keys())

    # trajectories
    if kapture.Trajectories in kapture_loadable_data:
        trajectories_file_path = csv_file_paths[kapture.Trajectories]
        logger.debug(f'loading trajectories {trajectories_file_path} ...')
        assert sensor_ids is not None
        kapture_data.trajectories = trajectories_from_file(trajectories_file_path, sensor_ids)

    _load_all_records(csv_file_paths, kapture_loadable_data, kapture_data)
    _load_features_and_desc_and_matches(data_dir_paths, kapture_dir_path, matches_pairs_file_path,
                                        kapture_loadable_data, kapture_data, tar_handlers)
    _load_points3d_and_observations(csv_file_paths, kapture_loadable_data, kapture_data)

    loading_elapsed = datetime.datetime.now() - loading_start
    logger.debug(f'Loaded in {loading_elapsed.total_seconds():.3f} seconds from "{kapture_dir_path}"')
    return kapture_data


def get_sensor_ids_of_type(sensor_type: str, sensors: kapture.Sensors) -> Set[str]:
    """
    Get the sensors of a certain kapture type ('camera', 'lidar', ...)

    :param sensor_type: type of sensor
    :param sensors: sensors to process
    :return: sensors identifiers
    """
    return set([sensor_id
                for sensor_id in sensors.keys()
                if sensors[sensor_id].sensor_type == sensor_type])


def _load_all_records(csv_file_paths, kapture_loadable_data, kapture_data) -> None:
    """
    Loads all records from disk to the kapture in memory

    :param csv_file_paths: file paths of the CVS records files
    :param kapture_loadable_data: the records to load
    :param kapture_data: to the kapture object to load into
    """
    # records camera
    if kapture.RecordsCamera in kapture_loadable_data:
        records_camera_file_path = csv_file_paths[kapture.RecordsCamera]
        logger.debug(f'loading images {records_camera_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type(kapture.SENSOR_TYPE_CAMERA, kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_camera = records_camera_from_file(csv_file_paths[kapture.RecordsCamera], sensor_ids)
    # records depth
    if kapture.RecordsDepth in kapture_loadable_data:
        records_depth_file_path = csv_file_paths[kapture.RecordsDepth]
        logger.debug(f'loading depth {records_depth_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type(kapture.SENSOR_TYPE_DEPTH_CAM, kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_depth = records_depth_from_file(csv_file_paths[kapture.RecordsDepth], sensor_ids)
    # records lidar
    if kapture.RecordsLidar in kapture_loadable_data:
        records_lidar_file_path = csv_file_paths[kapture.RecordsLidar]
        logger.debug(f'loading lidar {records_lidar_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('lidar', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_lidar = records_lidar_from_file(records_lidar_file_path, sensor_ids)
    # records Wifi
    if kapture.RecordsWifi in kapture_loadable_data:
        records_wifi_file_path = csv_file_paths[kapture.RecordsWifi]
        logger.debug(f'loading wifi {records_wifi_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('wifi', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_wifi = records_wifi_from_file(records_wifi_file_path, sensor_ids)

    # records bluetooth
    if kapture.RecordsBluetooth in kapture_loadable_data:
        records_bluetooth_file_path = csv_file_paths[kapture.RecordsBluetooth]
        logger.debug(f'loading bluetooth {records_bluetooth_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('bluetooth', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_bluetooth = records_bluetooth_from_file(records_bluetooth_file_path, sensor_ids)

    # records GNSS
    if kapture.RecordsGnss in kapture_loadable_data:
        records_gnss_file_path = csv_file_paths[kapture.RecordsGnss]
        logger.debug(f'loading GNSS {records_gnss_file_path} ...')
        assert kapture_data.sensors is not None
        epsg_codes = {sensor_id: sensor.sensor_params[0]
                      for sensor_id, sensor in kapture_data.sensors.items()
                      if sensor.sensor_type == 'gnss'}
        if len(epsg_codes) > 0:
            kapture_data.records_gnss = records_gnss_from_file(records_gnss_file_path, set(epsg_codes.keys()))
        else:
            logger.warning('no declared GNSS sensors: all GNSS data will be ignored')

    # records Accelerometer
    if kapture.RecordsAccelerometer in kapture_loadable_data:
        records_accelerometer_file_path = csv_file_paths[kapture.RecordsAccelerometer]
        logger.debug(f'loading Accelerations {records_accelerometer_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('accelerometer', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_accelerometer = records_accelerometer_from_file(records_accelerometer_file_path,
                                                                             sensor_ids)
    # records Gyroscope
    if kapture.RecordsGyroscope in kapture_loadable_data:
        records_gyroscope_file_path = csv_file_paths[kapture.RecordsGyroscope]
        logger.debug(f'loading Gyroscope {records_gyroscope_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('gyroscope', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_gyroscope = records_gyroscope_from_file(records_gyroscope_file_path, sensor_ids)
    # records Magnetic
    if kapture.RecordsMagnetic in kapture_loadable_data:
        records_magnetic_file_path = csv_file_paths[kapture.RecordsMagnetic]
        logger.debug(f'loading Magnetic {records_magnetic_file_path} ...')
        assert kapture_data.sensors is not None
        sensor_ids = get_sensor_ids_of_type('magnetic', kapture_data.sensors)
        assert sensor_ids is not None
        kapture_data.records_magnetic = records_magnetic_from_file(records_magnetic_file_path, sensor_ids)


def list_features(kapture_type: Type[Union[kapture.Keypoints,
                                           kapture.Descriptors,
                                           kapture.GlobalFeatures]],
                  kapture_dir_path: str) -> List[str]:
    """
    list available keypoints or descriptors or global features types for a given kapture

    :param kapture_type: whether to look inside keypoints, descriptors or global features
    :param kapture_dir_path: kapture top directory path
    """
    subfolders = (
        name
        for name in os.listdir(os.path.join(kapture_dir_path, kapture.io.features.FEATURES_DATA_DIRNAMES[kapture_type]))
        if os.path.isfile(get_feature_csv_fullpath(kapture_type, name, kapture_dir_path))
    )
    return list(subfolders)


def _load_features_and_desc_and_matches(data_dir_paths: dict, kapture_dir_path: str,
                                        matches_pairs_file_path: Optional[str],
                                        kapture_loadable_data: set, kapture_data: kapture.Kapture,
                                        tar_handlers: Optional[TarCollection] = None) -> None:
    """
    Loads features, descriptors, key points and matches from disk to the kapture in memory

    :param data_dir_paths: file paths of the data records files
    :param kapture_dir_path: kapture top directory path
    :param matches_pairs_file_path: text file in the csv format; where each line is image_name1, image_name2, score
    :param kapture_loadable_data: the data to load
    :param kapture_data: to the kapture object to load into
    :param tar_handler: collection of preloaded tar archives
    """

    # features
    image_filenames = set(image_name
                          for _, _, image_name in
                          kapture.flatten(kapture_data.records_camera)) \
        if kapture_data.records_camera is not None else set()
    # keypoints
    if kapture.Keypoints in kapture_loadable_data:
        logger.debug(f'loading keypoints {data_dir_paths[kapture.Keypoints]} ...')
        assert kapture_data.records_camera is not None
        keypoints_list = list_features(kapture.Keypoints, kapture_dir_path)
        if len(keypoints_list) > 0:
            kapture_data.keypoints = {}
            for keypoints_type in keypoints_list:
                kapture_data.keypoints[keypoints_type] = keypoints_from_dir(keypoints_type,
                                                                            kapture_dir_path,
                                                                            image_filenames,
                                                                            tar_handlers)
    # descriptors
    if kapture.Descriptors in kapture_loadable_data:
        logger.debug(f'loading descriptors {data_dir_paths[kapture.Descriptors]} ...')
        assert kapture_data.records_camera is not None
        descriptors_list = list_features(kapture.Descriptors, kapture_dir_path)
        if len(descriptors_list) > 0:
            kapture_data.descriptors = {}
            for descriptors_type in descriptors_list:
                kapture_data.descriptors[descriptors_type] = descriptors_from_dir(descriptors_type,
                                                                                  kapture_dir_path,
                                                                                  image_filenames,
                                                                                  tar_handlers)
    # global_features
    if kapture.GlobalFeatures in kapture_loadable_data:
        logger.debug(f'loading global features {data_dir_paths[kapture.GlobalFeatures]} ...')
        assert kapture_data.records_camera is not None
        global_features_list = list_features(kapture.GlobalFeatures, kapture_dir_path)
        if len(global_features_list) > 0:
            kapture_data.global_features = {}
            for global_features_type in global_features_list:
                kapture_data.global_features[global_features_type] = global_features_from_dir(global_features_type,
                                                                                              kapture_dir_path,
                                                                                              image_filenames,
                                                                                              tar_handlers)
    # matches
    if kapture.Matches in kapture_loadable_data:
        logger.debug(f'loading matches {data_dir_paths[kapture.Matches]} ...')
        assert kapture_data.records_camera is not None
        keypoints_list = [name
                          for name in os.listdir(data_dir_paths[kapture.Matches])
                          if os.path.isdir(os.path.join(data_dir_paths[kapture.Matches], name))]
        if len(keypoints_list) > 0:
            kapture_data.matches = {}
            for keypoints_type in keypoints_list:
                kapture_data.matches[keypoints_type] = matches_from_dir(keypoints_type,
                                                                        kapture_dir_path,
                                                                        image_filenames,
                                                                        matches_pairs_file_path,
                                                                        tar_handlers)


def _load_points3d_and_observations(csv_file_paths, kapture_loadable_data, kapture_data: kapture.Kapture) -> None:
    # points3d
    if kapture.Points3d in kapture_loadable_data:
        points3d_file_path = csv_file_paths[kapture.Points3d]
        logger.debug(f'loading points 3d {points3d_file_path} ...')
        kapture_data.points3d = points3d_from_file(points3d_file_path)
    # observations
    if kapture.Observations in kapture_loadable_data:
        observations_file_path = csv_file_paths[kapture.Observations]
        logger.debug(f'loading observations {observations_file_path} ...')
        assert kapture_data.keypoints is not None
        assert kapture_data.points3d is not None
        kapture_data.observations = observations_from_file(observations_file_path, kapture_data.keypoints)


def get_all_tar_handlers(kapture_dir_path: str,  # noqa: C901: function a bit long but not too complex
                         mode: Union[str, Dict[Type, str]] = 'r',
                         skip_list: List[Type[Union[
                             kapture.Keypoints,
                             kapture.Descriptors,
                             kapture.GlobalFeatures,
                             kapture.Matches
                         ]]] = []) -> TarCollection:
    """
    Preloads all tars for the kapture data in kapture_dir_path. can be either read of append.

    :param kapture_dir_path: kapture top directory path
    :param mode: 'r' or 'a', defaults to 'r'
    :param skip_list: Input option for expert only. Skip the load of specified parts, defaults to []
    :return: collection of preloaded tar archives, don't forget to call close() when you're done with it
    """
    valid_modes = {'r', 'a'}
    if isinstance(mode, str):
        assert mode in valid_modes
    else:
        assert isinstance(mode, dict)
        for mode_t in mode.values():
            assert mode_t in valid_modes

    data_dir_paths = {dtype: path.join(kapture_dir_path, dir_name)
                      for dtype, dir_name in kapture.io.features.FEATURES_DATA_DIRNAMES.items()}
    kapture_loadable_data = {
        kapture_type
        for kapture_type in KAPTURE_TARABLE_TYPES
        if kapture_type not in skip_list and path.exists(data_dir_paths[kapture_type])
    }
    tar_collection = TarCollection()

    # keypoints
    if kapture.Keypoints in kapture_loadable_data:
        logger.debug(f'opening keypoints tars {data_dir_paths[kapture.Keypoints]} ...')
        keypoints_list = list_features(kapture.Keypoints, kapture_dir_path)
        if len(keypoints_list) > 0:
            if isinstance(mode, str):
                mode_t = mode
            elif kapture.Keypoints in mode:
                mode_t = mode[kapture.Keypoints]
            else:
                mode_t = 'r'
            for keypoints_type in keypoints_list:
                tarfile_path = get_feature_tar_fullpath(kapture.Keypoints, keypoints_type, kapture_dir_path)
                if path.isfile(tarfile_path):
                    tar_collection.keypoints[keypoints_type] = TarHandler(tarfile_path, mode_t)
    # descriptors
    if kapture.Descriptors in kapture_loadable_data:
        logger.debug(f'opening descriptors tars {data_dir_paths[kapture.Descriptors]} ...')
        descriptors_list = list_features(kapture.Descriptors, kapture_dir_path)
        if len(descriptors_list) > 0:
            if isinstance(mode, str):
                mode_t = mode
            elif kapture.Descriptors in mode:
                mode_t = mode[kapture.Descriptors]
            else:
                mode_t = 'r'
            for descriptors_type in descriptors_list:
                tarfile_path = get_feature_tar_fullpath(kapture.Descriptors, descriptors_type, kapture_dir_path)
                if path.isfile(tarfile_path):
                    tar_collection.descriptors[descriptors_type] = TarHandler(tarfile_path, mode_t)
    # global_features
    if kapture.GlobalFeatures in kapture_loadable_data:
        logger.debug(f'opening global_features tars {data_dir_paths[kapture.GlobalFeatures]} ...')
        global_features_list = list_features(kapture.GlobalFeatures, kapture_dir_path)
        if len(global_features_list) > 0:
            if isinstance(mode, str):
                mode_t = mode
            elif kapture.GlobalFeatures in mode:
                mode_t = mode[kapture.GlobalFeatures]
            else:
                mode_t = 'r'
            for global_features_type in global_features_list:
                tarfile_path = get_feature_tar_fullpath(kapture.GlobalFeatures, global_features_type, kapture_dir_path)
                if path.isfile(tarfile_path):
                    tar_collection.global_features[global_features_type] = TarHandler(tarfile_path, mode_t)
    # matches
    if kapture.Matches in kapture_loadable_data:
        logger.debug(f'opening matches tars {data_dir_paths[kapture.Matches]} ...')
        keypoints_list = [name
                          for name in os.listdir(data_dir_paths[kapture.Matches])
                          if os.path.isdir(os.path.join(data_dir_paths[kapture.Matches], name))]
        if len(keypoints_list) > 0:
            if isinstance(mode, str):
                mode_t = mode
            elif kapture.Matches in mode:
                mode_t = mode[kapture.Matches]
            else:
                mode_t = 'r'
            for keypoints_type in keypoints_list:
                tarfile_path = get_feature_tar_fullpath(kapture.Matches, keypoints_type, kapture_dir_path)
                if path.isfile(tarfile_path):
                    tar_collection.matches[keypoints_type] = TarHandler(tarfile_path, mode_t)
    return tar_collection



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import Set, Tuple


class Matches(Set[Tuple[str, str]]):
    """
    Image matches
    """
    @staticmethod
    def lexical_order(image_path1: str, image_path2: str):
        """ returns the given pair in lexicographic order."""
        assert isinstance(image_path1, str)
        assert isinstance(image_path2, str)
        if image_path1 < image_path2:
            return image_path1, image_path2
        else:
            return image_path2, image_path1

    def add(self, image_path1: str, image_path2: str):
        """
        Inserts the image pair to matches. Does not ensure lexicographic order.
        Makes sure the entry guaranty that, or call normalize() afterward.

        :param image_path1:
        :param image_path2:
        :return:
        """
        assert isinstance(image_path1, str)
        assert isinstance(image_path2, str)
        # image_path1, image_path2 = self.lexical_order(image_path1, image_path2)
        super().add((image_path1, image_path2))

    def normalize(self):
        """ enforce lexicographic order on all matches. """
        temporary_set = Matches()
        for image_path1, image_path2 in self:
            image_path1, image_path2 = self.lexical_order(image_path1, image_path2)
            temporary_set.add(image_path1, image_path2)
        self.clear()
        self.update(temporary_set)

    def __repr__(self):
        if len(self) == 0:
            representation = 'no matches'
        else:
            representation = '[\n'
            representation += ',\n'.join(f'\t({image_path1} , {image_path2})' for (image_path1, image_path2) in self)
            representation += '\n]'
        return representation



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import argparse
import logging
import os.path as path
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture
from kapture.io.records import TransferAction
from kapture.converter.opensfm.import_opensfm import import_opensfm


logger = logging.getLogger('opensfm')


def import_opensfm_command_line():
    """
    Imports openSfM to kapture.
    """
    parser = argparse.ArgumentParser(description='Imports openSfM to kapture')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    ####################################################################################################################
    parser.add_argument('-i', '--input', '--opensfm', required=True,
                        help='path to OpenSfM root directory.')
    parser.add_argument('-o', '-k', '--output', '--kapture', required=True,
                        help='output directory where to save kapture files.')
    parser.add_argument('--transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], choose among: '
                             f'{", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete kapture if already exists.')
    parser.add_argument('--keypoints-type', default='HessianAffine', help='kapture keypoints type.')
    parser.add_argument('--descriptors-type', default='HOG', help='kapture descriptors type.')
    args = parser.parse_args()
    ####################################################################################################################
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug('\\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()
        if k != 'command'))

    args.input = path.normpath(path.abspath(args.input))
    args.output = path.normpath(path.abspath(args.output))

    import_opensfm(args.input, args.output, args.force, args.transfer, args.keypoints_type, args.descriptors_type)


if __name__ == '__main__':
    import_opensfm_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Convert kapture data in version 1.0 to version 1.1 inplace
"""

import logging
import os
import os.path as path
import argparse
from typing import Optional
import shutil
# import numpy as np like in kapture.io.csv
# so that types written as "np.float32" are understood by read_old_image_features_csv
import numpy as np  # noqa: F401

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.features
import kapture.io.csv
from kapture.utils.upgrade import upgrade_1_0_to_1_1_inplace

logger = logging.getLogger('upgrade_1_0_to_1_1_inplace')


def upgrade_1_0_to_1_1_inplace_command_line() -> None:
    """
    Convert kapture data in version 1.0 to version 1.1 inplace.
    """
    parser = argparse.ArgumentParser(
        description='convert kapture data in version 1.0 to version 1.1')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture directory')
    parser.add_argument('--keypoints-type', default=None,
                        help='types of keypoints.')
    parser.add_argument('--descriptors-type', default=None, help='types of descriptors.')
    parser.add_argument('--descriptors-metric-type', default='L2', help='types of descriptors.')
    parser.add_argument('--global-features-type', default=None,
                        help='types of global features.')
    parser.add_argument('--global-features-metric-type', default='L2', help='types of descriptors.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    upgrade_1_0_to_1_1_inplace(args.input,
                               args.keypoints_type, args.descriptors_type, args.global_features_type,
                               args.descriptors_metric_type, args.global_features_metric_type)


if __name__ == '__main__':
    upgrade_1_0_to_1_1_inplace_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import numpy as np
from dataclasses import dataclass, astuple, asdict, fields
from typing import Dict, List, Set, Tuple, TypeVar, Union

"""
Records contains sensor recordings.
Each recording (eg. an image) is an entry in Records, with a device ID and a timestamp.
There is to kind of Records :
 - Files: big recorded data (from camera, lidar, depth camera, ...) are stored in separated binary files.
 - Array: small recorded data (from wifi, gnss) are directly stored in kapture text files.

"""


########################################################################################################################
T = TypeVar('T')  # Declare generic type variable


class RecordsBase(Dict[int, Dict[str, T]]):
    """
    brief: Records
            records[timestamp][sensor_id] = <Record>
            or
            records[timestamp, sensor_id] = <Record>
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, T], T]):
        """
        Inserts/changes a record.

        :param key: can be timestamp or (timestamp, sensor_id)
        :param value: can be respectively a dict{sensor_id: DataRecords} or DataRecords
        """
        # enforce type checking
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            if not isinstance(value, self.record_type):
                raise TypeError(f'invalid record type of {type(value)} (expect {self.record_type})')
            self.setdefault(timestamp, {})[device_id] = value
        elif isinstance(key, int):
            # key is a timestamp
            timestamp = key
            if not isinstance(value, dict):
                raise TypeError('invalid value for data (expect dict)')
            if not all(isinstance(k, str) for k in value.keys()):
                raise TypeError('invalid device_id')
            if not all(isinstance(k, self.record_type) for k in value.values()):
                raise TypeError(f'invalid value for record (expect all {self.record_type})')
            super(RecordsBase, self).__setitem__(timestamp, value)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __getitem__(self, key: Union[int, Tuple[int, str]]) -> Union[Dict[str, T], T]:
        """
        Returns a single record for a (timestamp, sensor) or all records for a timestamp.

        :param key: can be timestamp or (timestamp, sensor_id)
        :return: return respectively a dict{sensor_id: DataRecords} or DataRecords
        """
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(RecordsBase, self).__getitem__(timestamp)[device_id]
        elif isinstance(key, int):
            # key is a timestamp
            return super(RecordsBase, self).__getitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __delitem__(self, key: Union[int, Tuple[int, str]]):
        """
        Delete an item or items.

        :param key: can be timestamp or (timestamp, sensor_id)
        """
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            super(RecordsBase, self).__getitem__(timestamp).__delitem__(device_id)
            if len(super(RecordsBase, self).__getitem__(timestamp)) == 0:
                # Cleaning upper level
                super(RecordsBase, self).__delitem__(timestamp)
        elif isinstance(key, int):
            # key is a timestamp
            super(RecordsBase, self).__delitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def key_pairs(self) -> List[Tuple[int, str]]:
        """
        Returns a list of (timestamp, device_id) contained in records.
        Those pairs can be used to access a single record data.

        :return: list of (timestamp, device_id)
        """
        return [
            (timestamp, sensor_id)
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        ]

    def data_list(self) -> List[T]:
        """
        Get the list of data stored for all sensor+timestamp values.

        :return: list of data
        """
        data_values = [data_value
                       for timestamp, sensors in self.items()
                       for sensor_id, data_value in sensors.items()]
        return data_values

    @property
    def sensors_ids(self) -> Set[str]:
        """
        :return: the set of unique sensors identifiers in the records data
        """
        return set(
            sensor_id
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        )

    def __contains__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(RecordsBase, self).__contains__(timestamp) and device_id in self[timestamp]
        elif isinstance(key, int):
            return super(RecordsBase, self).__contains__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __repr__(self) -> str:
        # [timestamp, sensor_id] = str
        lines = [f'[ {timestamp:010}, {sensor_id:5}] = {data}'
                 for timestamp, sensors in self.items()
                 for sensor_id, data in sensors.items()]
        return '\n'.join(lines)


class RecordsFilePath(RecordsBase[str]):
    """
    Brief: base class for records pointing to a data file (eg. camera, lidar).
    """
    record_type = str

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, str], str]):
        """ see RecordsBase.__setitem__ """
        if isinstance(key, tuple):
            if not isinstance(value, str):
                raise TypeError('invalid data')
        elif isinstance(key, int):
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            if not all(isinstance(v, str) for v in value.values()):
                raise TypeError('invalid data')
        super(RecordsFilePath, self).__setitem__(key, value)


class RecordsCamera(RecordsFilePath):
    """
    Camera records
    """
    pass


class RecordsDepth(RecordsFilePath):
    """
    Depth map records
    """
    dtype = np.float32


class RecordsLidar(RecordsFilePath):
    """
    Lidar records
    """
    pass


# Record Array #########################################################################################################
# @dataclass
class RecordArray:
    def __post_init__(self):
        # force cast to expected types
        for field in fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):
                setattr(self, field.name, field.type(value))

    def astuple(self):
        return astuple(self)

    def asdict(self):
        return asdict(self)

    @classmethod
    def fields(cls):
        return fields(cls)


class RecordsArray(RecordsBase[T]):
    """
    brief: Records
            records[timestamp][sensor_id] = RecordArray
            or
            records[timestamp, sensor_id] = RecordArray
    """
    pass


# wifi recordings is made of dict of signals (with signal strength).
@dataclass
class RecordWifiSignal(RecordArray):
    frequency: int
    rssi: float
    ssid: str = ''
    scan_time_start: int = 0
    scan_time_end: int = 0


class RecordWifi(dict):
    def __setitem__(self, bssid: str, data: RecordWifiSignal):
        if not isinstance(bssid, str):
            raise TypeError(f'{bssid} is not expected type str.')
        if not isinstance(data, RecordWifiSignal):
            raise TypeError(f'{data} is not expected type RecordWifiHotspot.')
        super().__setitem__(bssid, data)


class RecordsWifi(RecordsArray[RecordWifi]):
    """
    brief: Records wifi
            records[timestamp][sensor_id] =  <RecordWifi> = {bssid: RecordWifiHotspot}
            or
            records[timestamp, sensor_id] = <RecordWifi>
    """
    record_type = RecordWifi


# bluetooth recordings is made of dict of bt devices fingerprints (with signal strength).
@dataclass
class RecordBluetoothSignal(RecordArray):
    rssi: float
    name: str = ''


class RecordBluetooth(dict):
    def __setitem__(self, address: str, data: RecordBluetoothSignal):
        if not isinstance(address, str):
            raise TypeError(f'{address} is not expected type str.')
        if not isinstance(data, RecordBluetoothSignal):
            raise TypeError(f'{data} is not expected type RecordBluetoothDevice.')
        super().__setitem__(address, data)


class RecordsBluetooth(RecordsArray[RecordBluetooth]):
    """
    brief: Records wifi
            records[timestamp][sensor_id] = <RecordBluetooth> = {address: <RecordBluetoothDevice>}
            or
            records[timestamp, sensor_id] = <RecordBluetooth>
    """
    record_type = RecordBluetooth


# gnss recordings
@dataclass
class RecordGnss(RecordArray):
    x: float
    y: float
    z: float
    utc: int
    dop: float = 0.


class RecordsGnss(RecordsArray[RecordGnss]):
    record_type = RecordGnss


# Accelerometer recordings
@dataclass
class RecordAccelerometer(RecordArray):
    x_accel: float
    y_accel: float
    z_accel: float


class RecordsAccelerometer(RecordsArray[RecordAccelerometer]):
    record_type = RecordAccelerometer


# Gyroscope recordings
@dataclass
class RecordGyroscope(RecordArray):
    x_speed: float
    y_speed: float
    z_speed: float


class RecordsGyroscope(RecordsArray[RecordGyroscope]):
    record_type = RecordGyroscope


# Magnetic field recordings
@dataclass
class RecordMagnetic(RecordArray):
    x_strength: float
    y_strength: float
    z_strength: float


class RecordsMagnetic(RecordsArray[RecordMagnetic]):
    record_type = RecordMagnetic
